import atexit
import hashlib
import logging
import os
import pickle
import time

import dask
import dask.array as da
import numpy as np
import xarray as xr

try:
    from multiprocessing.shared_memory import ShareableList, SharedMemory
except ImportError:
    ShareableList, SharedMemory = None, None

try:
    import sparse
except ImportError:
    sparse = None

__GLOBAL_MEMORY_ARRAYS = {}
__GLOBAL_MEMORY_LISTS = {}

logger = logging.getLogger("sharrow.shared_memory")


def si_units(x, kind="B", digits=3, shift=1000):
    #       nano micro milli    kilo mega giga tera peta exa  zeta yotta
    tiers = ["n", "µ", "m", "", "K", "M", "G", "T", "P", "E", "Z", "Y"]

    tier = 3
    try:
        sign = "-" if x < 0 else ""
    except TypeError:
        # x is not a number, just return it
        return x
    x = abs(x)
    if x > 0:
        while x > shift and tier < len(tiers):
            x /= shift
            tier += 1
        while x < 1 and tier >= 0:
            x *= shift
            tier -= 1
    return f"{sign}{round(x,digits)} {tiers[tier]}{kind}"


def _hexhash(t, size=10, prefix="sharr"):
    h = hashlib.blake2b(str(t).encode(), digest_size=size).hexdigest()
    return prefix + h


def release_shared_memory(key=None):
    if key:
        v = __GLOBAL_MEMORY_ARRAYS.pop(key, None)
        if v:
            v.close()
            v.unlink()
        v = __GLOBAL_MEMORY_LISTS.pop(key, None)
        if v:
            v.shm.close()
            v.shm.unlink()
    else:
        while __GLOBAL_MEMORY_ARRAYS:
            k, v = __GLOBAL_MEMORY_ARRAYS.popitem()
            v.close()
            try:
                v.unlink()
            except FileNotFoundError:
                pass
        while __GLOBAL_MEMORY_LISTS:
            k, v = __GLOBAL_MEMORY_LISTS.popitem()
            v.shm.close()
            try:
                v.shm.unlink()
            except FileNotFoundError:
                pass


atexit.register(release_shared_memory)


def persist():
    atexit.unregister(release_shared_memory)


def create_shared_memory_array(key, size):
    logger.info(f"create_shared_memory_array({key}, )")

    if key.startswith("memmap:"):
        backing = key
    else:
        backing = "shared_memory"

    if backing == "shared_memory":
        logger.info(f"create_shared_memory_array:{key} ({si_units(size)})")
        h = _hexhash(f"sharrow__{key}")
        if h in __GLOBAL_MEMORY_ARRAYS:
            raise FileExistsError(f"sharrow_shared_memory_array:{key}")
        try:
            result = SharedMemory(
                name=h,
                create=True,
                size=size,
            )
        except FileExistsError:
            raise FileExistsError(f"sharrow_shared_memory_array:{key}") from None
        __GLOBAL_MEMORY_ARRAYS[key] = result
        return result

    if backing.startswith("memmap:"):
        logger.info(f"create_shared_memory_array:{key} ({backing}, {si_units(size)})")
        mmap_filename = backing[7:]
        if os.path.isfile(mmap_filename):
            result = np.memmap(
                mmap_filename,
                mode="r+",
            )
            if result.size != size:
                raise ValueError(f"file size mismatch, want {size} found {result.size}")
            return result
        else:
            return np.memmap(
                mmap_filename,
                mode="w+",
                shape=size,
            )

    raise ValueError(f"unknown backing {backing!r}")


def open_shared_memory_array(key, mode="r+"):
    logger.info(f"open_shared_memory_array:{key}")
    if key.startswith("memmap:"):
        backing = key
    else:
        backing = "shared_memory"
    if backing == "shared_memory":
        h = _hexhash(f"sharrow__{key}")
        try:
            result = SharedMemory(
                name=h,
                create=False,
            )
        except FileNotFoundError:
            raise FileNotFoundError(f"sharrow_shared_memory_array:{key}") from None
        else:
            logger.info(
                f"shared memory array from ephemeral memory, {si_units(result.size)}"
            )
            return result

    if backing.startswith("memmap:"):
        mmap_filename = backing[7:]
        result = np.memmap(
            mmap_filename,
            mode=mode,
            shape=None,
        )
        logger.info(f"shared memory array from memmap, {si_units(result.size)}")
        return result

    raise ValueError(f"unknown backing {backing!r}")


def create_shared_list(content, key):
    if key.startswith("memmap:"):
        mmap_filename = key[7:]
        os.makedirs(os.path.dirname(mmap_filename), exist_ok=True)
        with open(mmap_filename + ".meta.pkl", mode="wb") as f:
            pickle.dump(content, f)
    else:
        logger.info(f"create_shared_list:{key}")
        h = _hexhash(f"sharrow__list__{key}")
        if h in __GLOBAL_MEMORY_LISTS:
            raise FileExistsError(f"sharrow_shared_memory_list:{key}")
        try:
            result = ShareableList(
                content,
                name=h,
            )
        except FileExistsError:
            raise FileExistsError(f"sharrow_shared_memory_list:{key}") from None
        __GLOBAL_MEMORY_LISTS[key] = result
        return result


def read_shared_list(key):
    if key.startswith("memmap:"):
        mmap_filename = key[7:]
        with open(mmap_filename + ".meta.pkl", mode="rb") as f:
            return pickle.load(f)
    else:
        logger.info(f"read_shared_list:{key}")
        try:
            sl = ShareableList(name=_hexhash(f"sharrow__list__{key}"))
        except FileNotFoundError:
            raise FileNotFoundError(f"sharrow_shared_memory_list:{key}") from None
        else:
            return sl


def get_shared_list_nbytes(key):
    if key.startswith("memmap:"):
        return 0
    h = _hexhash(f"sharrow__list__{key}")
    try:
        shm = SharedMemory(name=h, create=False)
    except FileNotFoundError:
        raise FileNotFoundError(f"sharrow_shared_memory_list:{key}") from None
    else:
        return shm.size


def delete_shared_memory_files(key):
    if key.startswith("memmap:"):
        logger.info(f"delete_shared_memory_files:{key}")
        mmap_filename = key[7:]
        if os.path.isfile(mmap_filename):
            os.unlink(mmap_filename)
        if os.path.isfile(mmap_filename + ".meta.pkl"):
            os.unlink(mmap_filename + ".meta.pkl")


@xr.register_dataset_accessor("shm")
class SharedMemDatasetAccessor:
    _parent_class = xr.Dataset

    def __init__(self, xarray_obj):
        self._obj = xarray_obj
        self._shared_memory_key_ = None
        self._shared_memory_objs_ = []
        self._shared_memory_owned_ = False

    def _repr_html_(self):
        html = self._obj._repr_html_()
        html = html.replace("xarray.Dataset", "xarray.Dataset.shm")
        return html

    def __repr__(self):
        r = self._obj.__repr__()
        r = r.replace("xarray.Dataset", "xarray.Dataset.shm")
        return r

    def release_shared_memory(self):
        """Release shared memory allocated to this Dataset."""
        release_shared_memory(self._shared_memory_key_)

    @staticmethod
    def delete_shared_memory_files(key):
        delete_shared_memory_files(key)

    def to_shared_memory(
        self,
        key=None,
        mode="r+",
        _dupe=True,
        dask_scheduler="threads",
        pre_init=False,
        load=True,
    ):
        """
        Load this Dataset into shared memory.

        The returned Dataset object references the shared memory and is the
        "owner" of this data.  When this object is destroyed, the data backing
        it may also be freed, which can result in a segfault or other unfortunate
        condition if that memory is still accessed from elsewhere.

        Parameters
        ----------
        key : str
            An identifying key for this shared memory.  Use the same key
            in `from_shared_memory` to recreate this Dataset elsewhere.
        mode : {‘r+’, ‘r’, ‘w+’, ‘c’}, optional
            This method returns a copy of the Dataset in shared memory.
            If memmapped, that copy can be opened in various modes.
            See numpy.memmap() for details.
        dask_scheduler : str, default 'threads'
            The scheduler to use when loading dask arrays into shared memory.
            Typically, this is "threads" for multithreaded reads or "synchronous"
            for single-threaded reads. See dask.compute() for details.
        pre_init : bool, default False
            If True, the shared memory buffer will be pre-initialized with zeros.
            This is generally not necessary, but can be useful for debugging.
        load : bool, default True
            If True, load the data into shared memory immediately, using dask.
            If False, defer loading until later. Deferred tasks are stored in
            the `shm.tasks` attribute of the resulting Dataset object, but do not
            necessarily need to be run if data can be loaded using alternative
            methods (e.g. `sharrow.dataset.reload_from_omx_3d`).

        Returns
        -------
        Dataset
        """
        logger.info(f"sharrow.Dataset.to_shared_memory({key})")
        if key is None:
            import random

            key = random.randbytes(4).hex()
        self._shared_memory_key_ = key
        self._shared_memory_owned_ = False
        self._shared_memory_objs_ = []

        wrappers = []
        sizes = []
        names = []
        position = 0

        def emit(k, a, is_coord):
            nonlocal names, wrappers, sizes, position
            if sparse is not None and isinstance(a.data, sparse.GCXS):
                logger.debug(f"preparing sparse array {a.name}")
                wrappers.append(
                    {
                        "sparse": True,
                        "dims": a.dims,
                        "name": a.name,
                        "attrs": a.attrs,
                        "dtype": a.dtype,
                        "shape": a.shape,
                        "coord": is_coord,
                        "nbytes": a.data.nbytes,
                        "position": position,
                        "data.nbytes": a.data.data.nbytes,
                        "indices.nbytes": a.data.indices.nbytes,
                        "indptr.nbytes": a.data.indptr.nbytes,
                        "data.dtype": a.data.data.dtype,
                        "indices.dtype": a.data.indices.dtype,
                        "indptr.dtype": a.data.indptr.dtype,
                    }
                )
                a_nbytes = a.data.nbytes
            else:
                logger.debug(f"preparing dense array {a.name}")
                wrappers.append(
                    {
                        "dims": a.dims,
                        "name": a.name,
                        "attrs": a.attrs,
                        "dtype": a.dtype,
                        "shape": a.shape,
                        "coord": is_coord,
                        "nbytes": a.nbytes,
                        "position": position,
                    }
                )
                a_nbytes = a.nbytes

            sizes.append(a_nbytes)
            names.append(k)
            position += a_nbytes

        for k, a in self._obj.coords.items():
            emit(k, a, True)
        for k in self._obj.variables:
            if k in names:
                continue
            a = self._obj[k]
            emit(k, a, False)

        mem = create_shared_memory_array(key, size=position)

        logger.debug("declaring shared memory buffer")
        if key.startswith("memmap:"):
            buffer = memoryview(mem)
        else:
            buffer = mem.buf

        if pre_init:
            logger.debug("pre-initializing shared memory buffer")
            # gross init with all zeros
            buffer[:] = b"\0" * len(buffer)

        tasks = []
        task_names = []
        for w in wrappers:
            _is_sparse = w.get("sparse", False)
            _size = w["nbytes"]
            _name = w["name"]
            _pos = w["position"]
            a = self._obj[_name]
            if _is_sparse:
                logger.info(f"running load task: {_name} ({si_units(_size)})")
                ad = a.data
                _size_d = w["data.nbytes"]
                _size_i = w["indices.nbytes"]
                _size_p = w["indptr.nbytes"]
                mem_arr_d = np.ndarray(
                    shape=(_size_d // ad.data.dtype.itemsize,),
                    dtype=ad.data.dtype,
                    buffer=buffer[_pos : _pos + _size_d],
                )
                mem_arr_i = np.ndarray(
                    shape=(_size_i // ad.indices.dtype.itemsize,),
                    dtype=ad.indices.dtype,
                    buffer=buffer[_pos + _size_d : _pos + _size_d + _size_i],
                )
                mem_arr_p = np.ndarray(
                    shape=(_size_p // ad.indptr.dtype.itemsize,),
                    dtype=ad.indptr.dtype,
                    buffer=buffer[
                        _pos + _size_d + _size_i : _pos + _size_d + _size_i + _size_p
                    ],
                )
                mem_arr_d[:] = ad.data[:]
                mem_arr_i[:] = ad.indices[:]
                mem_arr_p[:] = ad.indptr[:]
            else:
                logger.debug(f"preparing load task: {_name} ({si_units(_size)})")
                mem_arr = np.ndarray(
                    shape=a.shape, dtype=a.dtype, buffer=buffer[_pos : _pos + _size]
                )
                if isinstance(a, xr.DataArray) and isinstance(a.data, da.Array):
                    tasks.append(da.store(a.data, mem_arr, lock=False, compute=False))
                    task_names.append(_name)
                else:
                    mem_arr[:] = a[:]
        if tasks and load:
            self.tasks = tasks
            self.task_names = task_names
            self.run_tasks(dask_scheduler=dask_scheduler)

        if key.startswith("memmap:"):
            mem.flush()

        logger.info("storing metadata in shared memory")
        create_shared_list(
            [pickle.dumps(self._obj.attrs)] + [pickle.dumps(i) for i in wrappers], key
        )
        result = type(self).from_shared_memory(key, own_data=mem, mode=mode)
        if tasks and not load:
            # attach incompleted tasks to the result
            result.shm.tasks = tasks
            result.shm.task_names = task_names
        result.shm._buffer = buffer
        result.shm._position = position
        return result

    def run_tasks(self, dask_scheduler="threads"):
        """Run any deferred dask tasks."""
        if not hasattr(self, "tasks"):
            return
        else:
            tasks = self.tasks
        if not hasattr(self, "task_names"):
            task_names = ["untitled" for _ in tasks]
        else:
            task_names = self.task_names
        t = time.time()
        logger.info(f"running {len(tasks)} dask data load tasks")
        if dask_scheduler == "synchronous":
            for task, task_name in zip(tasks, task_names):
                logger.info(f"running load task: {task_name}")
                dask.compute(task, scheduler=dask_scheduler)
        else:
            dask.compute(tasks, scheduler=dask_scheduler)
        logger.info(f"completed tasks in {time.time() - t:.3f} seconds")
        del self.tasks

    @property
    def shared_memory_key(self):
        try:
            return self._shared_memory_key_
        except AttributeError:
            raise ValueError("this dataset is not in shared memory") from None

    @classmethod
    def from_shared_memory(cls, key, own_data=False, mode="r+"):
        """
        Connect to an existing Dataset in shared memory.

        Parameters
        ----------
        key : str
            An identifying key for this shared memory.  Use the same key
            in `from_shared_memory` to recreate this Dataset elsewhere.
        own_data : bool or memmap, default False
            The returned Dataset object references the shared memory but is
            not the "owner" of this data unless this flag is set.  Pass a
            memmap to reuse an existing buffer.

        Returns
        -------
        Dataset
        """
        import pickle

        from xarray import DataArray

        _shared_memory_objs_ = []

        shr_list = read_shared_list(key)
        try:
            _shared_memory_objs_.append(shr_list.shm)
        except AttributeError:
            # for memmap, list is loaded from pickle, not shared ram
            pass

        if own_data and own_data is not True:
            mem = own_data
            own_data = True
        else:
            mem = open_shared_memory_array(key, mode=mode)
        _shared_memory_objs_.append(mem)
        if key.startswith("memmap:"):
            buffer = memoryview(mem)
        else:
            buffer = mem.buf

        content = {}

        for _n, w in enumerate(shr_list):
            if _n == 0:
                continue
            t = pickle.loads(w)
            shape = t.pop("shape")
            dtype = t.pop("dtype")
            name = t.pop("name")
            coord = t.pop("coord", False)  # noqa: F841
            position = t.pop("position")
            nbytes = t.pop("nbytes")
            is_sparse = t.pop("sparse", False)
            if is_sparse:
                if sparse is None:
                    raise ImportError("sparse is not installed")
                _size_d = t.pop("data.nbytes")
                _size_i = t.pop("indices.nbytes")
                _size_p = t.pop("indptr.nbytes")
                _dtype_d = t.pop("data.dtype")
                _dtype_i = t.pop("indices.dtype")
                _dtype_p = t.pop("indptr.dtype")

                mem_arr_d = np.ndarray(
                    _size_d // _dtype_d.itemsize,
                    dtype=_dtype_d,
                    buffer=buffer[position : position + _size_d],
                )
                mem_arr_i = np.ndarray(
                    _size_i // _dtype_i.itemsize,
                    dtype=_dtype_i,
                    buffer=buffer[position + _size_d : position + _size_d + _size_i],
                )
                mem_arr_p = np.ndarray(
                    _size_p // _dtype_p.itemsize,
                    dtype=_dtype_p,
                    buffer=buffer[
                        position + _size_d + _size_i : position
                        + _size_d
                        + _size_i
                        + _size_p
                    ],
                )
                mem_arr = sparse.GCXS(
                    (
                        mem_arr_d,
                        mem_arr_i,
                        mem_arr_p,
                    ),
                    shape=shape,
                    compressed_axes=(0,),
                )
            else:
                mem_arr = np.ndarray(
                    shape, dtype=dtype, buffer=buffer[position : position + nbytes]
                )
            content[name] = DataArray(mem_arr, **t)

        obj = cls._parent_class(content)
        obj.attrs = pickle.loads(shr_list[0])
        obj.shm._shared_memory_key_ = key
        obj.shm._shared_memory_owned_ = own_data
        obj.shm._shared_memory_objs_ = _shared_memory_objs_
        return obj

    @property
    def shared_memory_size(self):
        """Int : Size (in bytes) in shared memory, raises ValueError if not shared."""
        try:
            return sum(i.size for i in self._shared_memory_objs_)
        except AttributeError:
            raise ValueError("this dataset is not in shared memory") from None

    @property
    def is_shared_memory(self):
        """Bool : Whether this Dataset is in shared memory."""
        try:
            return sum(i.size for i in self._shared_memory_objs_) > 0
        except AttributeError:
            return False

    @staticmethod
    def preload_shared_memory_size(key):
        """
        Compute the size in bytes of a shared Dataset without actually loading it.

        Parameters
        ----------
        key : str
            The identifying key for this shared memory.

        Returns
        -------
        int
        """
        memsize = 0
        try:
            n = get_shared_list_nbytes(key)
        except FileNotFoundError:
            pass
        else:
            memsize += n
        try:
            mem = open_shared_memory_array(key, mode="r")
        except FileNotFoundError:
            pass
        else:
            memsize += mem.size
        return memsize

    def __getattr__(self, item):
        return getattr(self._obj, item)

    def __getitem__(self, item):
        return self._obj[item]



def example(extract='m', estimate=False):
    import larch

    larch.__version__

    d = larch.examples.MTC(format="dataset")
    d

    m = larch.Model(d, compute_engine="numba")

    from larch import PX, P, X

    for a in [2, 3]:
        m.utility_co[a] = +P("hhinc#2,3") * X("hhinc")
    for a in [4, 5, 6]:
        m.utility_co[a] = +P(f"hhinc#{a}") * X("hhinc")

    d.dc.alts_mapping

    for a, name in d.dc.alts_mapping.items():
        if a == 1:
            continue
        m.utility_co[a] += (
            +P("ASC_" + name)
            + P("vehbywrk_" + name) * X("vehbywrk")
            + P("wkempden_" + name) * X("wkempden")
        )

    m.utility_ca = (
        +PX("totcost")
        + P("motorized_time") * X("(altid <= 4) * tottime")
        + P("nonmotorized_time") * X("(altid > 4) * tottime")
        + P("motorized_ovtbydist") * X("(altid <= 4) * ovtt/dist")
    )

    m.availability_ca_var = "avail"
    m.choice_ca_var = "chose"

    m.title = "MTC Example 14, Work Zone Density"

    m.choice_avail_summary()

    m.set_cap(20)

    assert m.compute_engine == "numba"
    if not estimate:
        if isinstance(extract, str):
            return locals()[extract]
        else:
            _locals = locals()
            return [_locals.get(i) for i in extract]
    result = m.maximize_loglike(stderr=True, method="bhhh")
    m.calculate_parameter_covariance()
    m.loglike()

    m.parameter_summary()

    m.ordering = (
        (
            "LOS",
            ".*cost.*",
            ".*time.*",
            ".*dist.*",
        ),
        (
            "Income",
            "hhinc.*",
        ),
        ("Ownership", "vehbywrk.*"),
        ("Zonal", "wkcbd.*"),
        (
            "ASCs",
            "ASC.*",
        ),
    )

    m.parameter_summary()

    m.estimation_statistics()

    # TEST
    revealed_x = dict(zip(m.pnames, result.x))

    # TEST
    from pytest import approx

    expected_x = {
        "ASC_Bike": -1.597604389138021,
        "ASC_SR2": -1.604852502224575,
        "ASC_SR3+": -3.2123038625161646,
        "ASC_Transit": 0.4190578710080788,
        "ASC_Walk": -0.04100863602733044,
        "hhinc#2,3": -0.001774790620723634,
        "hhinc#4": -0.007075606155487522,
        "hhinc#5": -0.01118298552433892,
        "hhinc#6": -0.007903089907643455,
        "motorized_ovtbydist": -0.15748802513969926,
        "motorized_time": -0.029932168547457633,
        "nonmotorized_time": -0.04588150145362945,
        "totcost": -0.0028860643507599704,
        "vehbywrk_Bike": -0.7134510041177601,
        "vehbywrk_SR2": -0.40696239364371894,
        "vehbywrk_SR3+": -0.23731738368202027,
        "vehbywrk_Transit": -0.9946707857236002,
        "vehbywrk_Walk": -0.6812141773311737,
        "wkempden_Bike": 0.0010505639989477746,
        "wkempden_SR2": 0.0011364815897366538,
        "wkempden_SR3+": 0.0022153143196966783,
        "wkempden_Transit": 0.0026832716813030427,
        "wkempden_Walk": 0.0015105648018639885,
    }
    for k in expected_x:
        assert revealed_x[k] == approx(
            expected_x[k], 2e-2
        ), f"{k}, {revealed_x[k]/expected_x[k]}"
    if isinstance(extract, str):
        return locals()[extract]
    else:
        _locals = locals()
        return [_locals.get(i) for i in extract]



from __future__ import annotations

import heapq
import warnings
from collections import OrderedDict
from typing import TYPE_CHECKING, Literal

import networkx as nx
import numpy as np

from ..util.lazy import lazy

if TYPE_CHECKING:
    import pandas as pd


def add_split_key(d, k, v):
    if isinstance(k, (tuple, list)):
        if len(k) == 0:
            raise KeyError(f"empty {type(k)}")
        elif len(k) == 1:
            d[k[0]] = v
        else:
            if k[0] not in d:
                d[k[0]] = {}
            add_split_key(d[k[0]], k[1:], v)
    else:
        d[k] = v


def split_keys(d):
    out = {}
    for k, v in d.items():
        add_split_key(out, k, v)
    return out


class NestingTree(nx.DiGraph):
    node_dict_factory = OrderedDict
    adjlist_dict_factory = OrderedDict

    def copy(self, as_view=False):
        # copy operation needs to also copy the root_id and suggested_elemental_order
        result = super().copy(as_view=as_view)
        result._root_id = self._root_id
        result._suggested_elemental_order = self._suggested_elemental_order
        return result

    def __get__(self, instance, owner):
        # self : SubkeyStore
        # instance : instance of parent class that has `self` as a member, or None
        # owner : class of `instance`
        if instance is None:
            pass  # print("GRR: no instance")
            return self
        newself = getattr(instance, self.private_name, None)
        if newself is None:
            pass  # print(f"GRR No Current: {instance=} {owner=}")
            try:
                instance.initialize_graph()
            except ValueError:
                pass
            newself = getattr(instance, self.private_name, None)
        if newself is not None:
            newself._instance = instance
        pass  # print(f"GRR: get {instance=} {newself=}")
        return newself

    def __set__(self, instance, value):
        # self : NestingTree object
        # instance : instance of parent class that has `self` as a member
        # value : the new value that is trying to be assigned
        assert isinstance(value, NestingTree) or value is None
        if value is None:
            self.__delete__(instance)
            return
        t = value.copy()
        t._instance = instance
        setattr(instance, self.private_name, t)
        try:
            t._instance.mangle()
        except AttributeError:
            pass  # print(f"GRR: {err}")
        else:
            pass  # print(f"GRR Mangle: {instance}")

    def __delete__(self, instance):
        setattr(instance, self.private_name, None)
        try:
            instance.mangle()
        except AttributeError:
            pass  # print(f"GRR: {err}")
        else:
            pass  # print(f"GRR Mangle: {instance}")

    def __set_name__(self, owner, name):
        self.name = name
        self.private_name = "_private_" + name

    def touch(self):
        try:
            self._instance.mangle()
        except AttributeError:
            pass  # print("GRR: mangle failure")
        else:
            pass  # print("GRR: mangle ok")

    def __init__(self, *arg, root_id=0, suggested_elemental_order=(), **kwarg):
        if len(arg) and isinstance(arg[0], NestingTree):
            super().__init__(*arg, **kwarg)
            self._root_id = arg[0]._root_id
            if suggested_elemental_order != ():
                self._suggested_elemental_order = suggested_elemental_order
            else:
                self._suggested_elemental_order = arg[0]._suggested_elemental_order
        else:
            super().__init__(*arg, **kwarg)
            self._root_id = root_id
            self._suggested_elemental_order = suggested_elemental_order
        if self._root_id not in self.nodes:
            self.add_node(root_id, name="_root_", root=True)
        self._clear_caches()

    def __eq__(self, other):
        return (
            self._adj == other._adj
            and self._node == other._node
            and self._root_id == other._root_id
            and self._suggested_elemental_order == other._suggested_elemental_order
        )

    def suggest_elemental_order(self, order):
        self._suggested_elemental_order = tuple(j for j in order if j in self.nodes)

    @property
    def root_id(self):
        """Int : The code for the root node."""
        return self._root_id

    @root_id.setter
    def root_id(self, x):
        top_nests = list(self.successors(self._root_id))
        top_attrs = [self.edges[self._root_id, t] for t in top_nests]
        if self._root_id in self.nodes:
            self.remove_node(self._root_id)
        self._root_id = x
        if self._root_id not in self.nodes:
            self.add_node(self._root_id, name="_root_", root=True)
        for t, a in zip(top_nests, top_attrs):
            self.add_edge(self._root_id, t, **a, _clear_caches=False)
        self._clear_caches()

    def _clear_caches(self):
        NestingTree.topological_sorted.invalidate(self, "topological_sorted")
        NestingTree.topological_sorted_no_elementals.invalidate(
            self, "topological_sorted_no_elementals"
        )
        NestingTree.standard_sort.invalidate(self, "standard_sort")
        NestingTree.standard_sort.invalidate(self, "standard_slot_map")
        NestingTree.elementals.invalidate(self, "elementals")
        NestingTree.standard_competitive_edge_list.invalidate(
            self, "standard_competitive_edge_list"
        )
        NestingTree.standard_competitive_edge_list_2.invalidate(
            self, "standard_competitive_edge_list_2"
        )
        self._predecessor_slots = {}
        self._successor_slots = {}
        self.touch()

    def add_edge(self, u, v, implied=False, _clear_caches=True, **kwarg):
        """
        Add an edge between u and v.

        The nodes u and v will be automatically added if they are
        not already in the graph.

        Edge attributes can be specified with keywords.

        Parameters
        ----------
        u, v : int
                Nodes should be integer codes. The upstream node `u` is
                a nest or the root node.  Downsteam node `v` can be
                a nest or elemental alternative.
        implied : bool, default False
                Implied edges are for connection of otherwise unconnected
                nests to the root node.
        _clear_caches : bool, default True
        kwarg : keyword arguments, optional
                Edge data (or labels or objects) can be assigned using
                keyword arguments.
        """
        if not implied:
            drops = []
            for u_, v_, imp_ in self.in_edges(nbunch=[v], data="implied"):
                if imp_:
                    drops.append([u_, v_])
            for d in drops:
                super().remove_edge(*d)
        if _clear_caches:
            self._clear_caches()
        return super().add_edge(int(u), int(v), implied=implied, **kwarg)

    def _remove_edge_no_implied(self, u, v, *arg, **kwarg):
        result = super().remove_edge(u, v)
        self._clear_caches()
        return result

    def remove_edge(self, u, v, *arg, **kwarg):
        """
        Remove the edge between u and v.

        Parameters
        ----------
        u, v : int
                Remove the edge between nodes u and v.

        Raises
        ------
        NetworkXError
                If there is not an edge between u and v.
        """
        result = super().remove_edge(u, v)
        if self.in_degree(v) == 0 and v != self._root_id:
            self.add_edge(self._root_id, v, implied=True, _clear_caches=False)
        self._clear_caches()
        return result

    def add_node(
        self,
        code,
        *,
        children=(),
        parent=None,
        parents=None,
        phi_parameters=None,
        **kwarg,
    ):
        """
        Add a single node `code` and update node attributes.

        Parameters
        ----------
        code : int
                Although the generic networkx.DiGraph allows a node
                to be any hashable Python object except None, Larch
                assumes that node codes are integers.
        children : Collection
                A collection of other node codes that are the children
                of this new node.  Links will be created from this node
                to each child.
        parent : int, optional
                The parent of this new node. If not given, the root
                node is assumed to be the parent of this node, and an
                implied link is created.  This implied link is removed
                if the node is later made the child of some other node.
                If the parent is set explicitly, the link is *not*
                removed later.
        parents : Collection, optional
                Set multiple parent up-stream nodes.
        phi_parameters : Mapping
                Set phi parameters on graph links connecting to this
                node, used in network GEV models. The keys of this mapping
                indicate the node at the other end of the link, and the
                values are parameter names.
        kwarg : other keyword arguments, optional
                Set or change node attributes using key=value.
        """
        if parents is not None and parent is not None:
            raise TypeError("cannot give both parent and parents arguments")
        super().add_node(code, **kwarg)
        for child in children:
            self.add_edge(code, child, _clear_caches=False)
        if parent is not None:
            self.add_edge(parent, code, _clear_caches=False)
        elif parents is not None:
            for p in parents:
                self.add_edge(p, code, _clear_caches=False)
        else:
            if self.in_degree(code) == 0 and code != self._root_id:
                self.add_edge(self._root_id, code, implied=True, _clear_caches=False)
        if phi_parameters is not None:
            for k, parametername in phi_parameters.items():
                if (code, k) in self.edges:
                    self.edges[code, k]["parameter"] = str(parametername)
                elif (k, code) in self.edges:
                    self.edges[k, code]["parameter"] = str(parametername)
                else:
                    raise ValueError(
                        f"connected node {k} from phi_parameters not found"
                    )
        self._clear_caches()

    def new_node(self, *, code=None, **kwarg):
        """
        Add a new nesting node to this NestingTree.

        A new unique code is automatically created and returned by
        this method for creating new nests.

        All arguments must be given as keyword parameters.

        Parameters
        ----------
        parameter : str
                The name of the parameter to associate with this nest.
        children : Collection[int], optional
                The code numbers for the children of this nest.  These can be
                elemental alternatives or other nests.  If not given, no children
                will be defined initially, but they can be added later.
        parent : int, optional
                The code number for the parent of this nest.  If not given,
                the parent is implied as the root node, unless and until set
                to some other node.
        name : str, optional
                A human-readable name to associate with this nest.
        code : int, optional
                Use this code for the new nest. If this code already exists,
                a ValueError is raised.

        Returns
        -------
        int
                The new code for this nest.

        Raises
        ------
        ValueError
                If a new code is given but it already exists in this tree.
        """
        if code is None:
            proposed_code = len(self)
            while proposed_code in self:
                proposed_code += 1
        else:
            if code in self:
                raise ValueError(f"code {code} already exists in this tree")
            proposed_code = code
        self.add_node(proposed_code, **kwarg)
        return proposed_code

    def add_nodes(self, codes, *arg, parent=None, **kwarg):
        for code in codes:
            self.add_node(code, *arg, parent=parent, **kwarg)

    def remove_node(self, n):
        """
        Remove node n.

        Removes the node n, reconnecting all outedges to the head node
        of all inedges. Attempting to remove a non-existent node will
        raise an exception.

        Parameters
        ----------
        n : int
                A node in the graph
        """
        replace_edges = {k: self.edges[k].copy() for k in self.edges(n)}
        replace_heads = [k for k, _ in self.in_edges(n)]
        super().remove_node(n)
        for k, attrs in replace_edges.items():
            for h in replace_heads:
                super().add_edge(h, k[1], **attrs)
        self._clear_caches()

    @lazy
    def topological_sorted(self):
        return list(reverse_lexicographical_topological_sort(self))

    @lazy
    def topological_sorted_no_elementals(self):
        try:
            result = self.topological_sorted.copy()
        except nx.NetworkXUnfeasible:
            from networkx.algorithms.cycles import find_cycle

            try:
                cycle = find_cycle(self)
            except Exception:
                pass
            else:
                print("Found graph cycle:")
                print(list(cycle))
            raise

        # collect zero out-degree (elemental) codes in a set
        # to remove them in a batch, which is much faster than
        # removing them from the list one at a time.
        to_remove = set()
        for code, out_degree in self.out_degree:
            if not out_degree:
                to_remove.add(code)
        return [i for i in result if i not in to_remove]

    @lazy
    def standard_sort(self):
        return self.elementals + tuple(self.topological_sorted_no_elementals)

    def node_name(self, code):
        return self.nodes[code].get("name", str(code))

    @property
    def standard_sort_names(self):
        return [self.node_name(s) for s in self.standard_sort]

    def node_names(self):
        return {s: (self.node_name(s) or s) for s in self.standard_sort}

    def elemental_names(self):
        return {s: (self.node_name(s) or s) for s in self.elementals}

    @lazy
    def standard_slot_map(self):
        return {i: n for n, i in enumerate(self.standard_sort)}

    def predecessor_slots(self, code):
        if code in self._predecessor_slots:
            return self._predecessor_slots[code]
        s = np.empty(self.in_degree(code), dtype=np.int32)
        for n, i in enumerate(self.predecessors(code)):
            s[n] = self.standard_slot_map[i]
        self._predecessor_slots[code] = s
        return s

    def successor_slots(self, code):
        if code in self._successor_slots:
            return self._successor_slots[code]
        s = np.empty(self.out_degree(code), dtype=np.int32)
        for n, i in enumerate(self.successors(code)):
            s[n] = self.standard_slot_map[i]
        self._successor_slots[code] = s
        return s

    def __elementals_iter(self):
        for code, out_degree in self.out_degree:
            if not out_degree:
                yield code

    @lazy
    def elementals(self):
        result = []
        found = set()
        for e in self._suggested_elemental_order:
            if self.out_degree(e) == 0:
                result.append(e)
                found.add(e)
        for e in sorted(self.__elementals_iter()):
            if e not in found:
                result.append(e)
                found.add(e)
        return tuple(result)

    def n_elementals(self):
        return len(self.elementals)

    def n_intermediate_nests(self):
        return len(self.nodes) - self.n_elementals() - 1

    def elemental_descendants_iter(self, code):
        if not self.out_degree(code):
            yield code
            return
        all_d = nx.descendants(self, code)
        for dcode, dout_degree in self.out_degree(all_d):
            if not dout_degree:
                yield dcode

    def elemental_descendants(self, code):
        return [i for i in self.elemental_descendants_iter(code)]

    @property
    def n_edges(self):
        return self.number_of_edges()

    def edge_slot_arrays(self, alpha_locator=None):
        s = self.n_edges
        up = np.zeros(s, dtype=np.int32)
        dn = np.zeros(s, dtype=np.int32)
        first_visit = np.zeros(s, dtype=np.int32)
        alloc_slot = np.full_like(first_visit, -1)
        n = s
        first_visit_found = set()
        for upcode in reversed(self.standard_sort):
            upslot = self.standard_slot_map[upcode]
            for dnslot in reversed(self.successor_slots(upcode)):
                n -= 1
                up[n] = upslot
                dn[n] = dnslot
        for n in range(s):
            if dn[n] not in first_visit_found:
                first_visit[n] = 1
                first_visit_found.add(dn[n])
        if alpha_locator is not None:
            for n in range(s):
                alloc_slot[n] = alpha_locator.get(
                    (self.standard_sort[up[n]], self.standard_sort[dn[n]]), -1
                )
        return up, dn, first_visit, alloc_slot

    def nodes_with_successors_iter(self):
        for code, out_degree in self.out_degree:
            if out_degree:
                yield code

    def nodes_with_multiple_predecessors_iter(self):
        for code, in_degree in self.in_degree:
            if in_degree > 1:
                yield code

    @lazy
    def standard_competitive_edge_list(self):
        alphas = []
        for n in self.nodes_with_multiple_predecessors_iter():
            predecessors = sorted(self.predecessors(n))
            for k in predecessors:
                alphas.append((k, n))
        return alphas

    @lazy
    def standard_competitive_edge_list_2(self):
        alphas = []
        for n in self.nodes_with_multiple_predecessors_iter():
            predecessors = sorted(self.predecessors(n))
            alphas.append((predecessors, n))
        return alphas

    def __getstate__(self):
        attr = {}  # self.__dict__.copy()
        no_pickle = (
            "topological_sorted",
            "topological_sorted_no_elementals",
            "standard_sort",
            "standard_slot_map",
            #'_standard_elemental_sort',
            "elementals",
            "_predecessor_slots",
            "_successor_slots",
            "_touch",
            "node_dict_factory",
            "_instance",
        )
        for k, v in self.__dict__.items():
            if k not in no_pickle:
                attr[k] = v
        return attr

    def __setstate__(self, state):
        self.__dict__ = state.copy()
        self._predecessor_slots = {}
        self._successor_slots = {}

    def __xml__(self, use_viz=True, use_dot=True, output="svg", figsize=None, **format):
        viz = None
        dot = None
        if use_viz:
            try:
                import pygraphviz as viz
            except ImportError:
                if use_dot:
                    try:
                        import pydot as dot
                    except ImportError:
                        pass
        elif use_dot:
            try:
                import pydot as dot
            except ImportError:
                pass

        if viz is None and dot is None:
            if use_viz and use_dot:
                msg = "neither pydot nor pygraphviz modules are installed, unable to draw nesting tree"
            elif use_viz:
                msg = "pygraphviz module not installed, unable to draw nesting tree"
            elif use_dot:
                msg = "pydot module not installed, unable to draw nesting tree"
            else:
                msg = "no drawing module used, unable to draw nesting tree"
            warnings.warn(msg, stacklevel=2)
            raise NotImplementedError(msg)

        if viz is not None:
            existing_format_keys = list(format.keys())
            for key in existing_format_keys:
                if key.upper() != key:
                    format[key.upper()] = format[key]
            if "SUPPRESSGRAPHSIZE" not in format:
                if "GRAPHWIDTH" not in format:
                    format["GRAPHWIDTH"] = 6.5
                if "GRAPHHEIGHT" not in format:
                    format["GRAPHHEIGHT"] = 4
            if "UNAVAILABLE" not in format:
                format["UNAVAILABLE"] = True
            # x = XML_Builder("div", {'class':"nesting_graph larch_art"})
            # x.h2("Nesting Structure", anchor=1, attrib={'class':'larch_art_xhtml'})
            from io import BytesIO

            if "SUPPRESSGRAPHSIZE" not in format:
                G = viz.AGraph(
                    name="Tree",
                    directed=True,
                    size="{GRAPHWIDTH},{GRAPHHEIGHT}".format(**format),
                )
            else:
                G = viz.AGraph(name="Tree", directed=True)
            for n in self.nodes:
                nname = self.nodes[n].get("name", n)
                if nname == n:
                    G.add_node(
                        n,
                        label=f"<{nname}>",
                        style="rounded,solid",
                        shape="box",
                    )
                else:
                    G.add_node(
                        n,
                        label=f'<{nname} <FONT COLOR="#999999">({n})</FONT>>',
                        style="rounded,solid",
                        shape="box",
                    )
            _eG = G.add_subgraph(
                name="cluster_elemental",
                nbunch=self.elementals,
                color="#cccccc",
                bgcolor="#eeeeee",
                label="Elemental Alternatives",
                labelloc="b",
                style="rounded,solid",
            )
            # unavailable_nodes = set()
            # if format['UNAVAILABLE']:
            # 	if self.is_provisioned():
            # 		try:
            # 			for n, ncode in enumerate(self.alternative_codes()):
            # 				if np.sum(self.Data('Avail'),axis=0)[n,0]==0: unavailable_nodes.add(ncode)
            # 		except Exception: raise
            # 	try:
            # 		legible_avail = not isinstance(self.df.queries.avail, str)
            # 	except Exception:
            # 		legible_avail = False
            # 	if legible_avail:
            # 		for ncode,navail in self.df.queries.avail.items():
            # 			try:
            # 				if navail=='0': unavailable_nodes.add(ncode)
            # 			except Exception: raise
            # 	_eG.add_subgraph(
            #       name='cluster_elemental_unavailable', nbunch=unavailable_nodes,
            #       color='#bbbbbb', bgcolor='#dddddd',
            # 		label='Unavailable Alternatives', labelloc='b', style='rounded,solid'
            # 		)
            G.add_node(self.root_id, label="Root")
            up_nodes = set()
            down_nodes = set()
            for i, j in self.edges:
                G.add_edge(i, j)
                down_nodes.add(j)
                up_nodes.add(i)
            pyg_imgdata = BytesIO()
            try:
                G.draw(
                    pyg_imgdata, format=output, prog="dot"
                )  # write postscript in k5.ps with neato layout
            except ValueError as err:
                if "in path" in str(err):
                    warnings.warn(
                        str(err) + "; unable to draw nesting tree in report",
                        stacklevel=2,
                    )
                    raise NotImplementedError() from err
            if output == "svg":
                import xml.etree.ElementTree as ET

                ET.register_namespace("", "http://www.w3.org/2000/svg")
                ET.register_namespace("xlink", "http://www.w3.org/1999/xlink")
                return ET.fromstring(pyg_imgdata.getvalue().decode())
            else:
                raise NotImplementedError(f"output {output} with use_viz")
        else:
            pydot = dot

            # set Graphviz graph type
            if self.is_directed():
                graph_type = "digraph"
            else:
                graph_type = "graph"
            strict = nx.number_of_selfloops(self) == 0 and not self.is_multigraph()

            name = self.name
            graph_defaults = self.graph.get("graph", {})
            if name == "":
                P = pydot.Dot(
                    "", graph_type=graph_type, strict=strict, **graph_defaults
                )
            else:
                P = pydot.Dot(
                    f'"{name}"',
                    graph_type=graph_type,
                    strict=strict,
                    **graph_defaults,
                )
            try:
                P.set_node_defaults(**self.graph["node"])
            except KeyError:
                pass
            try:
                P.set_edge_defaults(**self.graph["edge"])
            except KeyError:
                pass

            cluster_elemental = pydot.Cluster(
                "elemental",
                style="rounded",
                bgcolor="lightgrey",
                color="white",
                rank="same",
                rankdir="LR",
            )

            for n, nodedata in self.nodes(data=True):
                str_nodedata = dict(
                    (k if k != "name" else "name_", '"' + str(v) + '"')
                    for k, v in nodedata.items()
                )

                if "parameter" in nodedata:
                    param_label = (
                        '<BR ALIGN="CENTER" />'
                        '<FONT COLOR="#999999" POINT-SIZE="9">'
                        "<I>{}</I>"
                        "</FONT>"
                    ).format(nodedata["parameter"])
                else:
                    param_label = ""

                if "name" in nodedata and n != self.root_id:
                    name = nodedata["name"]
                    str_nodedata["label"] = (
                        "<"
                        f'<FONT COLOR="#999999" POINT-SIZE="9">({n}) </FONT>'
                        f"{name}"
                        f"{param_label}>"
                    )

                # Default styling for nodes can have been overridden
                if n in self.elementals:
                    str_nodedata["style"] = str_nodedata.get("style", "filled")
                    str_nodedata["fillcolor"] = str_nodedata.get("fillcolor", "white")
                elif n == self.root_id:
                    str_nodedata["shape"] = str_nodedata.get("shape", "invhouse")
                else:
                    str_nodedata["style"] = str_nodedata.get("style", "rounded")
                    str_nodedata["shape"] = str_nodedata.get("shape", "rectangle")

                p = pydot.Node(str(n), **str_nodedata)
                P.add_node(p)
                if n in self.elementals:
                    cluster_elemental.add_node(p)

            P.add_subgraph(cluster_elemental)

            if self.is_multigraph():
                for u, v, key, edgedata in self.edges(data=True, keys=True):
                    str_edgedata = dict(
                        (k, str(v_)) for k, v_ in edgedata.items() if k != "key"
                    )
                    if v in self.elementals:
                        str_edgedata["constraint"] = "false"
                    edge = pydot.Edge(str(u), str(v), key=str(key), **str_edgedata)
                    P.add_edge(edge)

            else:
                for u, v, edgedata in self.edges(data=True):
                    str_edgedata = dict(
                        (k, '"' + str(v) + '"') for k, v in edgedata.items()
                    )
                    edge = pydot.Edge(str(u), str(v), **str_edgedata)
                    P.add_edge(edge)

            ###
            from xmle import Elem

            prog = None
            if output == "svg":
                import xml.etree.ElementTree as ET

                ET.register_namespace("", "http://www.w3.org/2000/svg")
                ET.register_namespace("xlink", "http://www.w3.org/1999/xlink")
            elif output == "png":
                prog = [P.prog, "-Gdpi=300"]
                if figsize is not None:
                    prog.append(rf"-Gsize={figsize[0]},{figsize[1]}\!")
                e = Elem.from_any(P.create(prog=prog, format=output, **format))
                e.attrib["dpi"] = (300, 300)
                return e
            return Elem.from_any(P.create(prog=prog, format=output, **format))

    def _repr_html_(self):
        from xmle import Elem

        try:
            xml = self.__xml__()
        except FileNotFoundError:
            return Elem("div", text="Graphviz not installed").tostring()
        else:
            x = Elem("div") << (xml)
        return x.tostring()

    def to_png(self, figsize=None, filename=None):
        """
        Output the graph visualization as a png.

        Parameters
        ----------
        figsize : 2-tuple, optional
                The (width, height) in inches.

        Returns
        -------
        xmle.Elem
        """
        result = self.__xml__(output="png", use_viz=False, figsize=figsize)
        if filename is not None:
            import base64

            if result.attrib["src"][:22] != "data:image/png;base64,":
                raise ValueError(
                    "problem decoding png:{}".format(result.attrib["src"][:22])
                )
            with open(filename, "wb") as fh:
                fh.write(base64.decodebytes(result.attrib["src"][22:].encode()))
        return result

    def partial_figure(
        self, including_nodes=None, source=None, *, n=None, n_at_level=3, n_expand=1
    ):
        """
        Generate a partial figure of the graph.

        Parameters
        ----------
        including_nodes : iterable or None
                An iterable containing node codes or names (or a mix).
        source : nodecode, optional
                All paths from this node to everything in `including_nodes` will be represented.
                Defaults to `root_id`.
        n : int
                If including_nodes is None, select this number of nodes randomly

        Returns
        -------
        Elem
        """
        if source is None:
            source = self.root_id
        from networkx.algorithms.simple_paths import all_simple_paths

        shows = set()

        if including_nodes is None and n is not None:
            including_nodes = sorted(np.random.choice(self.nodes, n, replace=False))

        if including_nodes is None and n_at_level is not None:
            import itertools
            from collections import deque

            q = deque([self.root_id])
            including_nodes = []
            while q:
                i = q.popleft()
                take = tuple(itertools.islice(self.successors(i), n_at_level))
                q.extend(take[:n_expand])
                including_nodes.extend(take)

        # Add every node in every path from the root to each `including_nodes`
        for each_node in including_nodes:
            if each_node in self.nodes:
                for i in all_simple_paths(self, source, each_node):
                    for j in i:
                        shows.add(j)
            else:
                for each_node_ in self.get_nodes_by_name(each_node):
                    for i in all_simple_paths(self, source, each_node_):
                        for j in i:
                            shows.add(j)
        s = self.subgraph(shows)
        return graph_to_figure(s)

    def get_nodes_by_name(self, name):
        result = [k for k, v in self.nodes(data=True) if v.get("name") == name]
        return result

    def subgraph_from(self, node):
        from collections import deque

        Q = deque([node])
        found = set()
        while len(Q):
            i = Q.popleft()
            if i not in found:
                found.add(i)
                Q.extend(self.successors(i))
        return NestingTree(self.subgraph(found), root_id=node)

    def stats_summarize(self):
        print("Graph Stats")
        print(f"  Overall: {len(self)} nodes")
        tier = [self.root_id]
        next_tier = list(self.successors(self.root_id))
        n = 0
        while len(next_tier):
            tier = next_tier
            n += 1
            print(f"   Tier {n}: {len(tier)} nodes")
            next_tier = list()
            for i in tier:
                next_tier.extend(self.successors(i))

    def node_slot_arrays(self, model, parameter_dict=None):
        """
        Return arrays for node parameters.

        Parameters
        ----------
        model : Model or dtype

        Returns
        -------
        muslots, start, num
        """
        if hasattr(model, "get_param_loc"):
            muslots = np.full(
                [
                    len(self),
                ],
                -1,
                dtype=np.int32,
            )
            for child, childcode in enumerate(self.standard_sort):
                # for parent in self.predecessor_slots(childcode):
                # 	alpha[parent, child] = 1
                pname = self.nodes[childcode].get("parameter", None)
                if pname is None:
                    muslots[child] = -1
                else:
                    muslots[child] = model.get_param_loc(pname)
        else:
            muslots = np.ones(
                [
                    len(self),
                ],
                dtype=model,
            )
            for child, childcode in enumerate(self.standard_sort):
                # for parent in self.predecessor_slots(childcode):
                # 	alpha[parent, child] = 1
                pname = self.nodes[childcode].get("parameter", None)
                if parameter_dict is not None and isinstance(pname, str):
                    pname = parameter_dict.get(pname, pname)
                if pname is not None:
                    muslots[child] = model(pname)
        num = np.zeros(len(self.nodes), dtype=np.int32)
        start = np.full(len(self.nodes), -1, dtype=np.int32)
        n = self.n_edges
        for upcode in reversed(self.standard_sort):
            upslot = self.standard_slot_map[upcode]
            for _dnslot in reversed(self.successor_slots(upcode)):
                n -= 1
                num[upslot] += 1
                start[upslot] = n
        return (
            muslots,
            start,
            num,
        )

    def _get_simple_mu_and_alpha(self, model, holdfast_invalidates=True):
        # alpha = np.zeros([len(self), len(self)], dtype=np.float64)
        mu = np.ones(
            [
                len(self),
            ],
            dtype=np.float64,
        )
        muslots = np.full(
            [
                len(self),
            ],
            -1,
            dtype=np.int32,
        )
        for child, childcode in enumerate(self.standard_sort):
            # for parent in self.predecessor_slots(childcode):
            # 	alpha[parent, child] = 1
            pname = self.nodes[childcode].get("parameter", None)
            mu[child] = model.get_value(pname, default=1.0)
            muslots[child] = model.get_param_loc(pname)
            if holdfast_invalidates:
                if model.pholdfast[muslots[child]]:
                    muslots[child] = -1

        s = self.n_edges
        up = np.zeros(s, dtype=np.int32)
        dn = np.zeros(s, dtype=np.int32)
        val = np.zeros(s, dtype=np.float64)
        num = np.zeros(len(self.nodes), dtype=np.int32)
        start = np.full(len(self.nodes), -1, dtype=np.int32)
        # first_visit = np.zeros(s, dtype=np.int32)
        n = s
        # first_visit_found = set()
        for upcode in reversed(self.standard_sort):
            upslot = self.standard_slot_map[upcode]
            for dnslot in reversed(self.successor_slots(upcode)):
                n -= 1
                up[n] = upslot
                dn[n] = dnslot
                num[upslot] += 1
                start[upslot] = n
                val[n] = 1 / len(
                    self.predecessor_slots(self.standard_sort[dnslot])
                )  # TODO make not always constant fraction
        # for n in range(s):
        # 	if dn[n] not in first_visit_found:
        # 		first_visit[n] = 1
        # 		first_visit_found.add(dn[n])

        return mu, muslots, up, dn, num, start, val

    def to_dict(self):
        return {
            "nodes": dict(self.nodes),
            "edges": split_keys(self.edges),
            "root_id": self.root_id,
            "suggested_elemental_order": self._suggested_elemental_order,
        }

    @classmethod
    def from_dict(cls, x):
        self = cls(
            root_id=x.get("root_id", None),
            suggested_elemental_order=x.get("suggested_elemental_order", ()),
        )
        nodes = x.get("nodes", {})
        for k, v in nodes.items():
            self.add_node(k, **v)
        edges = x.get("edges", {})
        for i, jv in edges.items():
            for j, v in jv.items():
                self.add_edge(i, j, **v)
        return self

    def as_arrays(self, model=np.float32, trim=True, parameter_dict=None):
        result = {}
        result["n_nodes"] = len(self)
        result["n_alts"] = n_alts = self.n_elementals()
        up, dn, first_visit, alloc_slot = self.edge_slot_arrays()
        result["edges_up"] = up
        result["edges_dn"] = dn
        result["edges_1st"] = first_visit
        result["edges_alloc"] = alloc_slot
        muslots, start, num = self.node_slot_arrays(
            model=model, parameter_dict=parameter_dict
        )
        if trim:
            muslots = muslots[n_alts:]
            start = start[n_alts:]
            num = num[n_alts:]
        result["mu_params"] = muslots
        result["start_slots"] = start
        result["len_slots"] = num
        return result

    def description(self, which: Literal["nodes", "edges"] = "nodes") -> pd.DataFrame:
        """
        Generate a DataFrame that describe this graph.

        Parameters
        ----------
        which : {'nodes','edges'}
            Which type of description to return.

        Returns
        -------
        DataFrame
            Describing nodes or edges
        """
        from itertools import islice

        import pandas as pd

        if which in ("edge", "edges", "link", "links"):
            ef = pd.DataFrame.from_dict(self.edges, orient="index")
            ef.index.names = ["up", "down"]
            return ef.fillna("")

        nf = pd.DataFrame.from_dict(self.nodes, orient="index")
        max_c = 8
        max_p = 8
        for n in nf.index:
            children = [e[1] for e in islice(self.out_edges(n), max_c)]
            if len(children) > max_c - 1:
                children[-1] = "..."
            if children:
                nf.loc[n, "children"] = ", ".join(str(c) for c in children)
        for n in nf.index:
            parents = [e[0] for e in islice(self.in_edges(n), max_p)]
            if len(parents) > max_p - 1:
                parents[-1] = "..."
            if parents:
                nf.loc[n, "parents"] = ", ".join(str(c) for c in parents)
        nf.drop(columns="root", errors="ignore", inplace=True)
        return nf.fillna("")


def graph_to_figure(graph, output_format="svg", **format):
    try:
        import pygraphviz as viz
    except ImportError:
        warnings.warn(
            "pygraphviz module not installed, unable to draw nesting tree", stacklevel=2
        )
        raise NotImplementedError(
            "pygraphviz module not installed, unable to draw nesting tree"
        ) from None
    existing_format_keys = list(format.keys())
    for key in existing_format_keys:
        if key.upper() != key:
            format[key.upper()] = format[key]
    if "SUPPRESSGRAPHSIZE" not in format:
        if "GRAPHWIDTH" not in format:
            format["GRAPHWIDTH"] = 6.5
        if "GRAPHHEIGHT" not in format:
            format["GRAPHHEIGHT"] = 4
    if "UNAVAILABLE" not in format:
        format["UNAVAILABLE"] = True
    # x = XML_Builder("div", {'class':"nesting_graph larch_art"})
    # x.h2("Nesting Structure", anchor=1, attrib={'class':'larch_art_xhtml'})
    from io import BytesIO

    if "SUPPRESSGRAPHSIZE" not in format:
        G = viz.AGraph(
            name="Tree",
            directed=True,
            size="{GRAPHWIDTH},{GRAPHHEIGHT}".format(**format),
        )
    else:
        G = viz.AGraph(name="Tree", directed=True)
    for n in graph.nodes:
        nname = graph.nodes[n].get("name", n)
        if nname == n:
            G.add_node(n, label=f"<{nname}>", style="rounded,solid", shape="box")
        else:
            G.add_node(
                n,
                label=f'<{nname} <FONT COLOR="#999999">({n})</FONT>>',
                style="rounded,solid",
                shape="box",
            )
    try:
        graph.elementals  # noqa: B018
    except AttributeError:
        pass
    else:
        _eG = G.add_subgraph(
            name="cluster_elemental",
            nbunch=graph.elementals,
            color="#cccccc",
            bgcolor="#eeeeee",
            label="Elemental Alternatives",
            labelloc="b",
            style="rounded,solid",
        )
    # unavailable_nodes = set()
    # if format['UNAVAILABLE']:
    # 	if self.is_provisioned():
    # 		try:
    # 			for n, ncode in enumerate(self.alternative_codes()):
    # 				if np.sum(self.Data('Avail'),axis=0)[n,0]==0: unavailable_nodes.add(ncode)
    # 		except Exception: raise
    # 	try:
    # 		legible_avail = not isinstance(self.df.queries.avail, str)
    # 	except Exception:
    # 		legible_avail = False
    # 	if legible_avail:
    # 		for ncode,navail in self.df.queries.avail.items():
    # 			try:
    # 				if navail=='0': unavailable_nodes.add(ncode)
    # 			except Exception: raise
    # 	eG.add_subgraph(
    #  	    name='cluster_elemental_unavailable',
    #  	    nbunch=unavailable_nodes,
    #  	    color='#bbbbbb', bgcolor='#dddddd',
    #  	    label='Unavailable Alternatives', labelloc='b', style='rounded,solid'
    #  	    )
    try:
        G.add_node(graph.root_id, label="Root")
    except AttributeError:
        pass
    up_nodes = set()
    down_nodes = set()
    for i, j in graph.edges:
        G.add_edge(i, j)
        down_nodes.add(j)
        up_nodes.add(i)
    pyg_imgdata = BytesIO()
    try:
        G.draw(
            pyg_imgdata, format=output_format, prog="dot"
        )  # write postscript in k5.ps with neato layout
    except ValueError as err:
        if "in path" in str(err):
            warnings.warn(
                str(err) + "; unable to draw nesting tree in report", stacklevel=2
            )
            raise NotImplementedError() from err
    from xmle import Elem

    if output_format == "svg":
        import xml.etree.ElementTree as ET

        ET.register_namespace("", "http://www.w3.org/2000/svg")
        ET.register_namespace("xlink", "http://www.w3.org/1999/xlink")
        result = ET.fromstring(pyg_imgdata.getvalue().decode())
    else:
        result = Elem(
            "span",
            attrib={"style": "color:red"},
            text=f"Unable to render output_format '{output_format}'",
        )
    x = Elem("div") << result
    return x


def reverse_lexicographical_topological_sort(G, key=None):
    """
    Generate nodes in reverse lexicographically topologically sorted order.

    A general topological sort is a nonunique permutation of the nodes such that
    an edge from u to v implies that u appears before v in the topological sort
    order.

    The lexicographical topological sort breaks ties by ordering according to
    node labels, so that the sorting becomes unique.

    Parameters
    ----------
    G : NetworkX digraph
            A directed acyclic graph (DAG)

    key : function, optional
            This function maps nodes to keys with which to resolve ambiguities in
            the sort order.  Defaults to the identity function.

    Returns
    -------
    iterable
            An iterable of node names in lexicographical topological sort order.

    Raises
    ------
    NetworkXError
            Topological sort is defined for directed graphs only. If the graph `G`
            is undirected, a :exc:`NetworkXError` is raised.

    NetworkXUnfeasible
            If `G` is not a directed acyclic graph (DAG) no topological sort exists
            and a :exc:`NetworkXUnfeasible` exception is raised.  This can also be
            raised if `G` is changed while the returned iterator is being processed

    RuntimeError
            If `G` is changed while the returned iterator is being processed.

    """
    if not G.is_directed():
        msg = "Topological sort not defined on undirected graphs."
        raise nx.NetworkXError(msg)

    if key is None:

        def key(node):
            return node

    nodeid_map = {n: i for i, n in enumerate(G)}

    def create_tuple(node):
        return key(node), nodeid_map[node], node

    outdegree_map = {v: d for v, d in G.out_degree() if d > 0}
    # These nodes have zero outdegree and ready to be returned.
    zero_outdegree = [create_tuple(v) for v, d in G.out_degree() if d == 0]
    heapq.heapify(zero_outdegree)

    while zero_outdegree:
        _, _, node = heapq.heappop(zero_outdegree)

        if node not in G:
            raise RuntimeError("Graph changed during iteration")
        for parent, _child in G.in_edges(node):
            try:
                outdegree_map[parent] -= 1
            except KeyError as e:
                raise RuntimeError("Graph changed during iteration") from e
            if outdegree_map[parent] == 0:
                heapq.heappush(zero_outdegree, create_tuple(parent))
                del outdegree_map[parent]

        yield node

    if zero_outdegree:
        msg = "Graph contains a cycle or graph changed during iteration"
        raise nx.NetworkXUnfeasible(msg)



import io
from itertools import zip_longest

try:
    from IPython.display import SVG
except ImportError:
    SVG = None

try:
    import pygraphviz as viz
except ImportError:
    viz = None


def make_graph(datatree, fontname="Arial", fontsize=12, ignore_joined_dims=True):
    if viz is None:
        raise ModuleNotFoundError("pygraphviz is not installed")

    small = f"{fontsize * .7:.1f}"
    g = viz.AGraph(rankdir="LR", strict=False, fontname=fontname, fontsize=fontsize)

    connected_nodes = set()
    for e in datatree._graph.edges:
        rel = datatree._get_relationship(e)
        connected_nodes.add(rel.parent_data)
        connected_nodes.add(rel.child_data)

    g_nodes = {}
    if len(datatree._graph.nodes):
        if datatree.root_node_name:
            if ignore_joined_dims:
                dims = [
                    i
                    for i in datatree.subspaces[datatree.root_node_name].dims
                    if not i.startswith("joined_")
                ]
            else:
                dims = list(datatree.subspaces[datatree.root_node_name].dims)
            g_nodes[datatree.root_node_name] = {"dims": dims, "vars": set()}
        for k in datatree._graph.nodes:
            if k == datatree.root_node_name or k not in connected_nodes:
                continue
            if ignore_joined_dims:
                dims = [
                    i for i in datatree.subspaces[k].dims if not i.startswith("joined_")
                ]
            else:
                dims = list(datatree.subspaces[k].dims)
            g_nodes[k] = {"dims": dims, "vars": set()}
    for e in datatree._graph.edges:
        rel = datatree._get_relationship(e)
        g_nodes[rel.parent_data]["vars"].add(rel.parent_name)

    def node_label_simple(k, v):
        dim_names = [f"<dim{i}>{i}" for i in v["dims"]]
        dim_defs = "|".join(dim_names)
        var_names = [f"<var{i}>{i}" for i in v["vars"]]
        var_defs = "|".join(var_names)
        out = f"<f0>{k}|{{{{Dimensions|{dim_defs}}}||{{Variables|{var_defs}}}}}"
        return out

    def node_label(k, v):
        cells = []
        if len(v["dims"]) > 0 and len(v["vars"]) > 0:
            cells.append(
                f"""
            <TR>
            <TD SIDES="L"><FONT COLOR="#999999" POINT-SIZE="{small}">DIMENSIONS</FONT></TD>
            <TD SIDES="R"><FONT COLOR="#999999" POINT-SIZE="{small}">VARIABLES</FONT></TD>
            </TR>
            """
            )
            for _d, _v in zip_longest(v["dims"], v["vars"], fillvalue=""):
                if _d == "":
                    cells.append(
                        f"""
                    <TR>
                    <TD BORDER="0"></TD>
                    <TD PORT="var{_v}" BGCOLOR="gray95">{_v}</TD>
                    </TR>
                    """
                    )
                elif _v == "":
                    cells.append(
                        f"""
                    <TR>
                    <TD PORT="dim{_d}" BGCOLOR="gray95">{_d}</TD>
                    <TD BORDER="0"></TD>
                    </TR>
                    """
                    )
                else:
                    cells.append(
                        f"""
                    <TR>
                    <TD PORT="dim{_d}" BGCOLOR="gray95">{_d}</TD>
                    <TD PORT="var{_v}" BGCOLOR="gray95">{_v}</TD>
                    </TR>
                    """
                    )
        elif len(v["dims"]) > 0 and len(v["vars"]) == 0:
            cells.append(
                f"""
            <TR>
            <TD SIDES="LR"><FONT COLOR="#999999" POINT-SIZE="{small}">DIMENSIONS</FONT></TD>
            </TR>
            """
            )
            for _d in v["dims"]:
                cells.append(
                    f"""
                <TR>
                <TD PORT="dim{_d}" BGCOLOR="gray95">{_d}</TD>
                </TR>
                """
                )

        out = f"""< <TABLE BORDER="0" CELLBORDER="1" CELLSPACING="0" >
        <TR><TD PORT="f0" COLSPAN="2" BGCOLOR="gray80"><B>{k}</B></TD></TR>
        {''.join(cells)}
        </TABLE> >"""
        return out

    if len(datatree._graph.nodes):
        for k, v in g_nodes.items():
            g.add_node(
                k,
                label=node_label(k, v),
                shape="plaintext",
                fontname=fontname,
                fontsize=fontsize,
            )

    for e in datatree._graph.edges:
        rel = datatree._get_relationship(e)
        g.add_edge(
            rel.parent_data,
            rel.child_data,
            key=repr(rel),
            tailport=f"var{rel.parent_name}",
            headport=f"dim{rel.child_name}",
            dir="forward",
            arrowhead="odiamond" if rel.indexing == "label" else "normal",
        )

    return g


def display_svg(graph):
    """
    Render a pygraphviz AGraph as SVG.

    Parameters
    ----------
    graph : pygraphviz.agraph.AGraph

    Returns
    -------
    IPython.core.display.SVG
    """
    pyg_imgdata = io.BytesIO()
    graph.draw(pyg_imgdata, format="svg", prog="dot")
    return SVG(pyg_imgdata.getvalue())



from __future__ import annotations

from enum import IntEnum
from functools import reduce

import numpy as np
import pandas as pd
import xarray as xr


class ArrayIsNotCategoricalError(TypeError):
    """The array is not an encoded categorical array."""


@xr.register_dataarray_accessor("cat")
class _Categorical:
    """Accessor for pseudo-categorical arrays."""

    __slots__ = ("dataarray",)

    def __init__(self, dataarray: xr.DataArray):
        self.dataarray = dataarray

    @property
    def categories(self):
        try:
            return self.dataarray.attrs["digital_encoding"]["dictionary"]
        except KeyError:
            raise ArrayIsNotCategoricalError() from None

    @property
    def ordered(self):
        return self.dataarray.attrs["digital_encoding"].get("ordered", True)

    def category_array(self) -> np.ndarray:
        arr = np.asarray(self.categories)
        if arr.dtype.kind == "O":
            arr = arr.astype(str)
        return arr

    def is_categorical(self) -> bool:
        return "dictionary" in self.dataarray.attrs.get("digital_encoding", {})

    def is_same_categories(self, other: xr.DataArray) -> bool:
        """Check if the categories of two categorical arrays are the same.

        Parameters
        ----------
        other : xr.DataArray or categorical accessor
            The other array to compare.
        """
        if not self.is_categorical():
            return False
        if isinstance(other, xr.DataArray):
            if not other.cat.is_categorical():
                return False
            if not np.array_equal(self.categories, other.cat.categories):
                return False
            if self.ordered != other.cat.ordered:
                return False
            return True
        elif isinstance(other, _Categorical):
            return self.is_same_categories(other.dataarray)
        else:
            raise TypeError(f"cannot compare categorical array with {type(other)}")


def _interpret_enum(e: type[IntEnum], value: int | str) -> IntEnum:
    """
    Convert a string or integer into an Enum value.

    The

    Parameters
    ----------
    e : Type[IntEnum]
        The enum to use in interpretation.
    value: int or str
        The value to convert.  Integer and simple string values are converted
        to their corresponding value.  Multiple string values can also be given
        joined by the pipe operator, in the style of flags (e.g. "Red|Octagon").
    """
    if isinstance(value, int):
        return e(value)
    return reduce(lambda x, y: x | y, [getattr(e, v) for v in value.split("|")])


def get_enum_name(e: type[IntEnum], value: int) -> str:
    """
    Get the name of an enum by value, or a placeholder name if not found.

    This allows for dummy placeholder names is the enum is does not contain
    all consecutive values between 0 and the maximum value, inclusive.

    Parameters
    ----------
    e : Type[IntEnum]
        The enum to use in interpretation.
    value : int
        The value for which to find a name.  If not found in `e`, this
        function will generate a new name as a string by prefixing `value`
        with an underscore.

    Returns
    -------
    str
    """
    result = e._value2member_map_.get(value, f"_{value}")
    try:
        return result.name
    except AttributeError:
        return result


def int_enum_to_categorical_dtype(e: type[IntEnum]) -> pd.CategoricalDtype:
    """
    Convert an integer-valued enum to a pandas CategoricalDtype.

    Parameters
    ----------
    e : Type[IntEnum]

    Returns
    -------
    pd.CategoricalDtype
    """
    max_enum_value = int(max(e))
    categories = [get_enum_name(e, i) for i in range(max_enum_value + 1)]
    return pd.CategoricalDtype(categories=categories)


def as_int_enum(
    s: pd.Series,
    e: type[IntEnum],
    dtype: type[np.integer] | None = None,
    categorical: bool = True,
) -> pd.Series:
    """
    Encode a pandas Series as categorical, consistent with an IntEnum.

    Parameters
    ----------
    s : pd.Series
    e : Type[IntEnum]
    dtype : Type[np.integer], optional
        Specific dtype to use for the code point encoding.  It is typically not
        necessary to give this explicitly as the function will automatically
        select the best (most efficient) bitwidth.
    categorical : bool, default True
        If set to false, the returned series will simply be integer encoded with
        no formal Categorical dtype.

    Returns
    -------
    pd.Series
    """
    min_enum_value = int(min(e))
    max_enum_value = int(max(e))
    assert min_enum_value >= 0
    if dtype is None:
        if max_enum_value < 256 and min_enum_value >= 0:
            dtype = np.uint8
        elif max_enum_value < 128 and min_enum_value >= -128:
            dtype = np.int8
        elif max_enum_value < 65536 and min_enum_value >= 0:
            dtype = np.uint16
        elif max_enum_value < 32768 and min_enum_value >= -32768:
            dtype = np.int16
        elif max_enum_value < 2_147_483_648 and min_enum_value >= -2_147_483_648:
            dtype = np.int32
        else:
            dtype = np.int64
    if not isinstance(s, pd.Series):
        s = pd.Series(s)
    result = s.apply(lambda x: _interpret_enum(e, x)).astype(dtype)
    if categorical:
        categories = [get_enum_name(e, i) for i in range(max_enum_value + 1)]
        result = pd.Categorical.from_codes(codes=result, categories=categories)
    return result


@pd.api.extensions.register_series_accessor("as_int_enum")
class _AsIntEnum:
    """
    Encode a pandas Series as categorical, consistent with an IntEnum.

    Parameters
    ----------
    s : pd.Series
    e : Type[IntEnum]
    dtype : Type[np.integer], optional
        Specific dtype to use for the code point encoding.  It is typically not
        necessary to give this explicitly as the function will automatically
        select the best (most efficient) bitwidth.
    categorical : bool, default True
        If set to false, the returned series will simply be integer encoded with
        no formal Categorical dtype.

    Returns
    -------
    pd.Series
    """

    def __init__(self, pandas_obj):
        self._obj = pandas_obj

    def __call__(
        self: pd.Series,
        e: type[IntEnum],
        dtype: type[np.integer] | None = None,
        categorical: bool = True,
    ):
        return as_int_enum(self._obj, e, dtype, categorical)



def example(extract='m', estimate=False):
    import larch

    larch.__version__

    d = larch.examples.MTC(format="dataset")
    d

    m = larch.Model(d, compute_engine="numba")

    from larch import PX, P, X

    for a in [2, 3]:
        m.utility_co[a] = +P("hhinc#2,3") * X("hhinc")
    for a in [4, 5, 6]:
        m.utility_co[a] = +P(f"hhinc#{a}") * X("hhinc")

    d.dc.alts_mapping

    for a, name in d.dc.alts_mapping.items():
        if a == 1:
            continue
        m.utility_co[a] += P("ASC_" + name) + P("vehbyadlt_" + name) * X("numveh/numadlt")

    m.utility_ca = (
        +PX("totcost")
        + P("motorized_time") * X("(altid <= 4) * tottime")
        + P("nonmotorized_time") * X("(altid > 4) * tottime")
        + P("motorized_ovtbydist") * X("(altid <= 4) * ovtt/dist")
    )

    m.availability_ca_var = "avail"
    m.choice_ca_var = "chose"

    m.title = "MTC Example 12, Autos per Adult"

    m.choice_avail_summary()

    m.set_cap(20)

    assert m.compute_engine == "numba"
    if not estimate:
        if isinstance(extract, str):
            return locals()[extract]
        else:
            _locals = locals()
            return [_locals.get(i) for i in extract]
    result = m.maximize_loglike(stderr=True, method="bhhh")
    m.calculate_parameter_covariance()
    m.loglike()

    m.parameter_summary()

    m.ordering = (
        (
            "LOS",
            ".*cost.*",
            ".*time.*",
            ".*dist.*",
        ),
        (
            "Income",
            "hhinc.*",
        ),
        ("Ownership", "vehbyadlt.*"),
        (
            "ASCs",
            "ASC.*",
        ),
    )

    m.parameter_summary()

    m.estimation_statistics()

    # TEST
    from pytest import approx

    revealed_x = dict(zip(m.pnames, result.x))
    assert result["loglike"] == approx(-3490.3579715315, rel=1e-5), result["loglike"]

    expected_x = {
        "ASC_Bike": -1.9627758221667733,
        "ASC_SR2": -1.536865403172976,
        "ASC_SR3+": -3.0235322387181176,
        "ASC_Transit": 1.0496197735578745,
        "ASC_Walk": -0.21728572066754234,
        "hhinc#2,3": -0.0013240180758255255,
        "hhinc#4": -0.004616956633813837,
        "hhinc#5": -0.011663606663415026,
        "hhinc#6": -0.00764683683920611,
        "motorized_ovtbydist": -0.18489944106276082,
        "motorized_time": -0.03803405138642473,
        "nonmotorized_time": -0.046633097590087214,
        "totcost": -0.004179712660535923,
        "vehbyadlt_Bike": -0.6418346308251331,
        "vehbyadlt_SR2": -0.594456791391658,
        "vehbyadlt_SR3+": -0.447408231859455,
        "vehbyadlt_Transit": -1.4092358161539453,
        "vehbyadlt_Walk": -0.7935735138713005,
    }
    for k in expected_x:
        assert revealed_x[k] == approx(
            expected_x[k], 0.03
        ), f"{k}, {revealed_x[k]/expected_x[k]}"
    if isinstance(extract, str):
        return locals()[extract]
    else:
        _locals = locals()
        return [_locals.get(i) for i in extract]



from __future__ import annotations

import os

from xmle import NumberedCaption, Reporter
from xmle import load_metadata as read_metadata

from . import examples
from ._version import version as _install_version
from .dataset import DataArray, Dataset, DataTree
from .examples import example, example_file
from .model import mixtures
from .model.basemodel import BaseModel
from .model.jaxmodel import Model
from .model.latent_class import LatentClass, MixedLatentClass
from .model.model_group import ModelGroup
from .model.param_core import ParameterBucket
from .model.saving import load_model
from .model.tree import NestingTree
from .shorts import PX, P, X
from .util import figures

# Get decorated version when in development
try:
    top_package = __import__(__name__.split(".")[0])
    import setuptools_scm

    __version__ = setuptools_scm.get_version(os.path.dirname(top_package.__path__[0]))
except Exception:
    __version__ = _install_version

__all__ = [
    "NumberedCaption",
    "Reporter",
    "read_metadata",
    "examples",
    "example",
    "example_file",
    "DataArray",
    "Dataset",
    "DataTree",
    "mixtures",
    "BaseModel",
    "Model",
    "ModelGroup",
    "LatentClass",
    "MixedLatentClass",
    "ParameterBucket",
    "load_model",
    "NestingTree",
    "PX",
    "P",
    "X",
    "figures",
    "versions",
]


def versions() -> dict:
    """Versions of Larch and its key dependencies."""
    import numba as nb
    import numpy as np
    import pandas as pd
    import sharrow as sh
    import xarray as xr

    return {
        "larch": __version__,
        "sharrow": sh.__version__,
        "numpy": np.__version__,
        "pandas": pd.__version__,
        "xarray": xr.__version__,
        "numba": nb.__version__,
    }



def example(extract='m', estimate=False):
    import numpy as np
    from pytest import approx

    import larch as lx

    import larch as lx

    varnames = [
        "price",
        "time",
        "conven",
        "comfort",
        "meals",
        "petfr",
        "emipp",
        "nonsig1",
        "nonsig2",
        "nonsig3",
    ]
    d = lx.examples.ARTIFICIAL()
    m = lx.Model(d)
    m.utility_ca = sum(lx.PX(i) for i in varnames)
    m.choice_ca_var = "choice"
    randvars_normal = ["meals", "petfr", "emipp"]

    m.mixtures = [lx.mixtures.Normal(k, f"sd.{k}") for k in randvars_normal]

    m.n_draws = 300
    m.seed = 42
    m.common_draws = True
    if not estimate:
        if isinstance(extract, str):
            return locals()[extract]
        else:
            _locals = locals()
            return [_locals.get(i) for i in extract]
    result = m.maximize_loglike(stderr=True)

    m2 = m.copy()
    m2.pvals = "init"
    m2.common_draws = False

    result2 = m2.maximize_loglike(stderr=True)

    m2.parameter_summary()

    from xlogit import MixedLogit

    df = d.to_dataframe().reset_index()
    varnames = [
        "price",
        "time",
        "conven",
        "comfort",
        "meals",
        "petfr",
        "emipp",
        "nonsig1",
        "nonsig2",
        "nonsig3",
    ]
    X = df[varnames].values
    y = df["choice"].values
    randvars = {"meals": "n", "petfr": "n", "emipp": "n"}
    alts = df["alt"]
    ids = df["id"]
    panels = None
    batch_size = 5000
    n_draws = 300

    np.random.seed(0)
    model = MixedLogit()
    model.fit(
        X,
        y,
        varnames,
        alts=alts,
        ids=ids,
        n_draws=n_draws,
        panels=panels,
        verbose=0,
        randvars=randvars,
        batch_size=batch_size,
    )

    model.summary()
    if isinstance(extract, str):
        return locals()[extract]
    else:
        _locals = locals()
        return [_locals.get(i) for i in extract]



from __future__ import annotations

import numpy as np
import pandas as pd
import pytest
from pytest import approx

import larch as lx
from larch import PX, P, X
from larch.exceptions import MissingDataError


def test_bad_avail_declaration():
    d = lx.examples.MTC(format="dataset")
    m = lx.Model(d)
    with pytest.raises(TypeError):
        m.unknown_attr = 123
    with pytest.raises(TypeError):
        m.availability_var = "avail"


def test_model_has_no_data():
    m = lx.Model()
    with pytest.raises(MissingDataError):
        m.maximize_loglike()


def test_model_with_altcode_zero():
    raw_data = pd.read_csv(lx.example_file("swissmetro.csv.gz")).rename_axis(
        index="CASEID"
    )
    ds = lx.Dataset.construct.from_idco(raw_data, alts={0: "Ticketed", 1: "GA"})
    m = lx.Model(ds, compute_engine="numba")
    m.title = "Swiss GA Travelcard (binary logit)"
    m.choice_co_code = "GA"
    m.utility_co[1] = (
        +P.intercept + P.age * X.AGE + P.male * X.MALE + P.income * X.INCOME
    )
    m.maximize_loglike()
    assert m._most_recent_estimation_result["loglike"] == approx(-4258.78631691331)
    assert m._most_recent_estimation_result["x"].to_dict() == approx(
        {
            "age": -0.27650147517278006,
            "income": -0.05339095646626831,
            "intercept": -0.4818265416235945,
            "male": -0.6078366425546569,
        }
    )


def test_model_choice_avail():
    d = lx.examples.MTC(format="dataset")
    m = lx.Model(d, compute_engine="numba")
    m.availability_ca_var = "avail"
    m.choice_ca_var = "chose"
    cas = m.choice_avail_summary()
    np.testing.assert_array_equal(cas["name"].values[:-1], d.coords["alt_names"].data)
    np.testing.assert_array_equal(
        cas["chosen"].values, [3637, 517, 161, 498, 50, 166, 5029]
    )
    np.testing.assert_array_equal(
        cas["available"].values[:-1], [4755, 5029, 5029, 4003, 1738, 1479]
    )


def test_model_choice_avail_weighted():
    d = lx.examples.MTC(format="dataset")
    m = lx.Model(d, compute_engine="numba")
    m.availability_ca_var = "avail"
    m.choice_ca_var = "chose"
    m.weight_co_var = "famtype"
    cas = m.choice_avail_summary()
    np.testing.assert_array_equal(cas["name"].values[:-1], d.coords["alt_names"].data)
    np.testing.assert_array_equal(
        cas["chosen unweighted"].values, [3637, 517, 161, 498, 50, 166, 5029]
    )
    np.testing.assert_array_equal(
        cas["available unweighted"].values[:-1], [4755, 5029, 5029, 4003, 1738, 1479]
    )
    np.testing.assert_array_almost_equal(
        cas["chosen weighted"].values,
        [5851.0, 2563.0, 8512.0, 758.0, 2760.0, 59820.0, 39376.0],
    )
    np.testing.assert_array_almost_equal(
        cas["available weighted"].values,
        [59820.0, 59820.0, 48023.0, 22370.0, 19273.0, 59820.0, np.nan],
    )


def test_model_gradient_check():
    d = lx.examples.MTC(format="dataset")
    m = lx.Model(d, compute_engine="numba")
    m.availability_ca_var = "avail"
    m.choice_ca_var = "chose"
    m.utility_co[2] = P("ASC_SR2") + P("hhinc#2") * X("hhinc")
    m.utility_co[3] = P("ASC_SR3P") + P("hhinc#3") * X("hhinc")
    m.utility_co[4] = P("ASC_TRAN") + P("hhinc#4") * X("hhinc")
    m.utility_co[5] = P("ASC_BIKE") + P("hhinc#5") * X("hhinc")
    m.utility_co[6] = P("ASC_WALK") + P("hhinc#6") * X("hhinc")
    m.utility_ca = PX("tottime") + PX("totcost")
    chk = m.check_d_loglike()
    assert chk.data.similarity.min() > 5
    dll = {
        "ASC_BIKE": -279.79999999999376,
        "ASC_SR2": -687.7000000000156,
        "ASC_SR3P": -1043.7000000000337,
        "ASC_TRAN": -380.78333333332864,
        "ASC_WALK": -113.65000000000244,
        "hhinc#2": -41179.541666666715,
        "hhinc#3": -60691.541666666686,
        "hhinc#4": -24028.374999999985,
        "hhinc#5": -17374.80833333334,
        "hhinc#6": -7739.108333333339,
        "totcost": 127397.53666666638,
        "tottime": -42104.202666666555,
    }
    assert chk.data.analytic.to_dict() == approx(dll)
    assert chk.data.finite_diff.to_dict() == approx(dll, rel=1e-5)



from __future__ import annotations

from collections.abc import Callable, Collection, Mapping
from io import BytesIO

import pandas
from xmle import Elem


def adjust_spines(ax, spines):
    for loc, spine in ax.spines.items():
        if loc in spines:
            # spine.set_position(('outward', 10))  # outward by 10 points
            # spine.set_smart_bounds(True)
            pass
        else:
            spine.set_color("none")  # don't draw spine
    # turn off ticks where there is no spine
    if "left" in spines:
        ax.yaxis.set_ticks_position("left")
    else:
        # no yaxis ticks
        ax.yaxis.set_ticks([])
    if "bottom" in spines:
        ax.xaxis.set_ticks_position("bottom")
    else:
        # no xaxis ticks
        ax.xaxis.set_ticks([])


def plot_as_svg_xhtml(
    fig,
    classname="figure",
    headerlevel=2,
    header=None,
    anchor=1,
    transparent=True,
    tooltip=None,
    bbox_extra_artists=None,
    pad_inches=0.1,
    **format,
):
    existing_format_keys = list(format.keys())
    for key in existing_format_keys:
        if key.upper() != key:
            format[key.upper()] = format[key]
    if "GRAPHWIDTH" not in format and "GRAPHHEIGHT" in format:
        format["GRAPHWIDTH"] = format["GRAPHHEIGHT"]
    if "GRAPHWIDTH" in format and "GRAPHHEIGHT" not in format:
        format["GRAPHHEIGHT"] = format["GRAPHWIDTH"] * 0.67
    import xml.etree.ElementTree as ET

    ET.register_namespace("", "http://www.w3.org/2000/svg")
    ET.register_namespace("xlink", "http://www.w3.org/1999/xlink")
    imgbuffer = BytesIO()
    fig.savefig(
        imgbuffer,
        dpi=None,
        facecolor="none",
        edgecolor="w",
        orientation="portrait",
        format="svg",
        transparent=transparent,
        bbox_inches="tight",
        pad_inches=pad_inches,
        bbox_extra_artists=bbox_extra_artists,
    )
    x = Elem("div", {"class": classname})
    if header:
        x.hn(headerlevel, header, anchor=anchor)
    x << ET.fromstring(imgbuffer.getvalue().decode())
    if tooltip is not None:
        x[0][1].insert(0, Elem("title", text=tooltip))
    return x


def line_graph(
    x,
    y,
    x_title=None,
    y_title=None,
    show_legend=None,
    **kwargs,
):
    """
    Generate a line graph.

    Parameters
    ----------
    x : array-like
    y : array-like or pandas.Series or pandas.DataFrame
    x_title : str
    y_title : str
    show_legend : bool, optional
    **kwargs
    """
    useful_legend = False

    if x_title is None:
        x_title = getattr(x, "name", None)

    from .._optional import pyplot as plt

    fig, ax = plt.subplots()

    if isinstance(y, pandas.DataFrame):
        for col in y.columns:
            ax.plot(x, y[col], label=col)
            useful_legend = True
    elif isinstance(y, pandas.Series):
        plt.plot(x, y.values, label=y.name)
        if y.name is not None:
            useful_legend = True
    elif isinstance(y, Collection) and all(isinstance(k, Callable) for k in y):
        for k in y:
            this_y = k(x)
            label_y = getattr(this_y, "name", None)
            ax.plot(x, this_y, label=label_y)
            useful_legend = useful_legend or (label_y is not None)
    elif isinstance(y, Mapping):
        for col, val in y.items():
            ax.plot(x, val, label=col)
        useful_legend = True
    else:
        ax.plot(x, y)

    if x_title is not None:
        ax.set_xlabel(x_title)

    if y_title is not None:
        ax.set_ylabel(y_title)

    if show_legend is None and useful_legend:
        ax.legend()
    elif show_legend:
        ax.legend()

    fig.tight_layout(pad=0.5)
    result = plot_as_svg_xhtml(fig, **kwargs)
    plt.close(fig)

    return result



from __future__ import annotations

import numpy as np
import pandas as pd
import pytest
from pytest import approx

import larch as lx
from larch import P, X
from larch.model.troubleshooting import PossibleOverspecificationError


@pytest.fixture(params=["numba", "jax"])
def ref_model(request):
    if request.param == "jax":
        pytest.importorskip("jax")
    m = lx.example(1)
    m.compute_engine = request.param
    m.graph.new_node(parameter="mu_motor", children=[1, 2, 3, 4], name="Motorized")
    m.graph.new_node(parameter="mu_nonmotor", children=[5, 6], name="Nonmotorized")
    m.unmangle()
    return m


def test_chosen_but_not_available_query(ref_model: lx.Model):
    m = ref_model

    # change first 3 observations to WALK
    m.dataset = m.dataset.assign(ch=m.dataset["ch"].copy())
    m.dataset["ch"][:3, :] = 0
    m.dataset["ch"][:3, -1] = 1

    # test just checking
    m, problems = m.doctor(repair_ch_av="?")
    assert len(problems) == 1
    assert "chosen_but_not_available" in problems
    ll = m.loglike_casewise()
    assert ll[0] < -9e9
    assert ll[1] < -9e9
    assert ll[2] < -9e9

    # test raising error
    with pytest.raises(ValueError):
        m, problems = m.doctor(repair_ch_av="!")

    # this problem included by default in doctor exceptions
    with pytest.raises(ValueError):
        m, problems = m.doctor()


def test_chosen_but_not_available_plus(ref_model: lx.Model):
    m = ref_model

    # change first 3 observations to WALK
    m.dataset = m.dataset.assign(ch=m.dataset["ch"].copy())
    m.dataset["ch"][:3, :] = 0
    m.dataset["ch"][:3, -1] = 1

    m, problems = m.doctor(repair_ch_av="+")
    assert (m.dataset["av"][:3, -1] == 1).all()
    ll = m.loglike_casewise()
    assert ll[:3] == approx([-1.79175947, -1.79175947, -1.60943791])


def test_chosen_but_not_available_minus(ref_model: lx.Model):
    m = ref_model

    # change first 3 observations to WALK
    m.dataset = m.dataset.assign(ch=m.dataset["ch"].copy())
    m.dataset["ch"][:3, :] = 0
    m.dataset["ch"][:3, -1] = 1

    m, problems = m.doctor(repair_ch_av="-", repair_noch_nzwt=None)
    assert (m.dataset["av"][:3, -1] == 0).all()
    assert (m.dataset["ch"][:3, -1] == 0).all()
    ll = m.loglike_casewise()
    assert ll[:3] == approx([0, 0, 0])
    assert m.loglike() == approx(-7304.995801563645)


def test_nothing_chosen_but_nonzero_weight(ref_model: lx.Model):
    m = ref_model
    m.weight_co_var = "famtype"

    # change first 3 observations to no choice
    m.dataset = m.dataset.assign(ch=m.dataset["ch"].copy())
    m.dataset["ch"][:3, :] = 0

    m, problems = m.doctor(repair_noch_nzwt="?")
    assert "nothing_chosen_but_nonzero_weight" in problems

    # this problem included by default in doctor exceptions
    with pytest.raises(ValueError):
        m, problems = m.doctor()


def test_nothing_chosen_but_nonzero_weight_minus(ref_model: lx.Model):
    m = ref_model
    m.weight_co_var = "famtype"

    # change first 3 observations to no choice
    m.dataset = m.dataset.assign(ch=m.dataset["ch"].copy())
    m.dataset["ch"][:3, :] = 0

    m, problems = m.doctor(repair_noch_nzwt="-")
    assert "nothing_chosen_but_nonzero_weight" in problems
    assert (m.dataset["wt"][:3] == 0).all()
    ll = m.loglike_casewise()
    assert ll[:3] == approx([0, 0, 0])


def test_nan_in_data_co(ref_model: lx.Model):
    m = ref_model

    # change first 3 observations to have NaN for income
    m.dataset = m.dataset.assign(co=m.dataset["co"].copy())
    m.dataset["co"][:3, 0] = np.nan
    m._rebuild_data_arrays()

    assert np.isnan(m.dataset["co"][:3, 0]).all()
    assert np.isnan(m.loglike(error_if_bad=False))
    with pytest.raises(ValueError, match="log likelihood is NaN"):
        m.loglike(error_if_bad=True)

    # test raising error
    with pytest.raises(ValueError):
        m, problems = m.doctor(repair_nan_data_co="!")

    # test just checking
    m, problems = m.doctor(repair_nan_data_co="?")
    assert len(problems) == 1
    assert "nan_data_co" in problems

    # this problem included by default in doctor exceptions
    with pytest.raises(ValueError):
        m, problems = m.doctor()


def test_low_variance_data_co(ref_model: lx.Model):
    m = ref_model

    # change first 3 observations to have NaN for income
    m.dataset = m.dataset.assign(co=m.dataset["co"].copy())
    m.dataset["co"][:, 0] = 42.0
    m._rebuild_data_arrays()

    # test raising error
    with pytest.raises(ValueError):
        m, problems = m.doctor(check_low_variance_data_co="!")

    # test just checking
    m, problems = m.doctor(check_low_variance_data_co="?")
    assert len(problems) == 1
    assert "low_variance_data_co" in problems

    # this problem included by default in doctor warnings not exceptions
    m, problems = m.doctor()
    assert "low_variance_data_co" in problems


def test_nan_in_weight(ref_model: lx.Model):
    m = ref_model
    m.weight_co_var = "famtype"

    # change some observations to have NaN for weight
    m.dataset = m.dataset.assign(wt=m.dataset["wt"].copy())
    m.dataset["wt"][3:9] = np.nan
    m._rebuild_data_arrays()

    assert np.isnan(m.dataset["wt"][3:9]).all()
    assert np.isnan(m.loglike(error_if_bad=False))
    with pytest.raises(ValueError, match=".* is NaN in 6 cases"):
        m.loglike(error_if_bad=True)

    # test raising error
    with pytest.raises(ValueError):
        m, problems = m.doctor(repair_nan_wt="!")

    # test just checking
    m, problems = m.doctor(repair_nan_wt="?")
    assert len(problems) == 1
    assert "nan_weight" in problems

    # this problem included by default in doctor exceptions
    with pytest.raises(ValueError):
        m, problems = m.doctor()


@pytest.mark.parametrize("compute_engine", ["numba", "jax"])
def test_chosen_but_zero_quantity(compute_engine):
    if compute_engine == "jax":
        pytest.importorskip("jax")

    hh, pp, tour, skims, emp = lx.example(200, ["hh", "pp", "tour", "skims", "emp"])
    base = lx.Dataset.construct.new_idca(tour.TOURID, skims.TAZ_ID)
    hh["INCOME_GRP"] = pd.qcut(hh.INCOME, 3)

    # set employment in a couple zones to zero
    emp.loc[2] = 0
    emp.loc[9] = 0

    tree = lx.DataTree(
        base=base,
        tour=tour.rename_axis(index="TOUR_ID"),
        hh=hh.set_index("HHID"),
        person=pp.set_index("PERSONID"),
        emp=emp,
        skims=lx.Dataset.construct.from_omx(skims),
        relationships=(
            "base.TAZ_ID @ emp.TAZ",
            "base.TOURID -> tour.TOUR_ID",
            "tour.HHID @ hh.HHID",
            "tour.PERSONID @ person.PERSONID",
            "hh.HOMETAZ @ skims.otaz",
            "base.TAZ_ID @ skims.dtaz",
        ),
    )

    m = lx.Model(datatree=tree, compute_engine=compute_engine)
    m.quantity_ca = (
        +P.EmpRetail_HighInc * X("RETAIL_EMP * (INCOME>50000)")
        + P.EmpNonRetail_HighInc * X("NONRETAIL_EMP") * X("INCOME>50000")
        + P.EmpRetail_LowInc * X("RETAIL_EMP") * X("INCOME<=50000")
        + P.EmpNonRetail_LowInc * X("NONRETAIL_EMP") * X("INCOME<=50000")
    )
    m.quantity_scale = P.Theta
    m.choice_co_code = "tour.DTAZ"

    # test raising error
    with pytest.raises(ValueError):
        m, problems = m.doctor(repair_ch_zq="!")

    # test just checking
    m, problems = m.doctor(repair_ch_zq="?")
    assert len(problems) == 1
    assert "chosen_but_zero_quantity" in problems

    the_problem = pd.DataFrame.from_records(
        np.rec.array(
            [(2, 528, "302, 413, 415"), (9, 557, "357, 373, 388")],
            dtype=[("TAZ_ID", "<i8"), ("n", "<i8"), ("example rows", "O")],
        )
    ).set_index("TAZ_ID")

    pd.testing.assert_frame_equal(problems.chosen_but_zero_quantity, the_problem)

    # this problem included by default in doctor exceptions
    with pytest.raises(ValueError):
        m, problems = m.doctor()

    # test that log likelihood error checking works, for numba
    # for jax there is clipping and no error is raised
    if compute_engine == "numba":
        with pytest.raises(ValueError, match="log likelihood is Inf"):
            m.loglike()


def test_overspec():
    m = lx.example(1)
    m.utility_co[1] = P.ASC_DA

    # test raising error
    with pytest.raises(PossibleOverspecificationError):
        m, problems = m.doctor(check_overspec="!")

    # test just checking
    m, problems = m.doctor(check_overspec="?")
    assert len(problems) == 1
    assert "overspecification" in problems

    diagnosis = problems["overspecification"].to_dict()["eigenvector"]
    key2s = []
    for key, val in diagnosis.items():
        assert key[0] == 0
        key2s.append(key[2])
        # eigenvalues near zero are not stable, don't check the value
        # assert key[1] == approx(2.8603410100913607e-05, rel=0.1)
        assert np.absolute(val) == approx(0.40825, rel=0.1)
    assert set(key2s) == {
        "ASC_BIKE",
        "ASC_DA",
        "ASC_SR2",
        "ASC_SR3P",
        "ASC_TRAN",
        "ASC_WALK",
    }

    # this problem included by default in doctor warnings, not an exception
    m, problems = m.doctor()
    assert "overspecification" in problems



from __future__ import annotations

import numba as nb
import numpy as np
import pandas as pd
from sharrow import Flow


@nb.njit(nogil=True)
def cascade_sum_1row(arr, dn_slots, up_slots, n_nodes, max_val=None):
    out = np.zeros(n_nodes, dtype=arr.dtype)
    out[: arr.size] = arr[:]
    for j in range(dn_slots.size):
        if max_val is None:
            out[up_slots[j]] += out[dn_slots[j]]
        else:
            k = out[up_slots[j]] + out[dn_slots[j]]
            out[up_slots[j]] = min(k, max_val)
    return out


def _raw_streamer(source, name, dtype):
    raw_data = source.root_dataset[name].to_numpy()
    if raw_data.dtype == dtype:
        func = lambda c: raw_data[c]
    else:
        func = lambda c: raw_data[c].astype(dtype)
    return nb.njit(func)


class OneHotStreamer:
    def __init__(self, key, coords):
        self.key = key
        self.coords = coords

    def compile(self, source, dtype):
        raw_codes = source._getitem(self.key).to_numpy()
        n_alts = source.n_alts
        coords = np.asarray(self.coords)

        def pseudo_flow(c):
            out = np.zeros(n_alts, dtype=dtype)
            this_code = raw_codes[c]
            for i in range(coords.size):
                if this_code == coords[i]:
                    out[i] = 1
                    break
            return out

        return nb.njit(pseudo_flow)


def get_data_streamers(model, default_dtype=np.float64):
    dtypes = {"av": np.int8, "avail_ca": np.int8}
    streamers = {}
    dataflows = model.dataflows
    for k, v in dataflows.items():
        if k in {"co", "wt"}:
            source = model.datatree.idco_subtree()
        else:
            source = model.datatree
        if isinstance(v, Flow):
            streamers[k] = source.reuse_streamer(v, dtype=dtypes.get(k, default_dtype))
            # streamers[k] = v.init_streamer(
            #     source=source,
            #     dtype=dtypes.get(k, default_dtype),
            #     closure=True,
            # )
        elif isinstance(v, str):
            streamers[k] = _raw_streamer(source, v, dtypes.get(k, default_dtype))
        elif isinstance(v, OneHotStreamer):
            streamers[k] = v.compile(source, dtypes.get(k, default_dtype))
        else:
            raise NotImplementedError
    return streamers


def init_streamer(
    model,
):
    from .numbamodel import (  # inside function to prevent circular import
        _numba_utility_to_loglike,
        quantity_from_data_ca,
        utility_from_data_ca,
        utility_from_data_co,
    )

    streamers = get_data_streamers(model, model.float_dtype)
    float_dtype = model.float_dtype
    n_alts = model.datatree.n_alts

    if "ca" in streamers:
        stream_ca = streamers["ca"].compiled
    else:
        stream_ca = nb.njit(lambda c: np.zeros((n_alts, 0), dtype=float_dtype))

    if "co" in streamers:
        stream_co = streamers["co"].compiled
    else:
        stream_co = nb.njit(lambda c: np.zeros((0,), dtype=float_dtype))

    if "avail_ca" in streamers:
        _stream_av_1 = streamers["avail_ca"]
        _stream_av = streamers["avail_ca"].compiled
        stream_av = nb.njit(lambda c: np.reshape(_stream_av(c), -1))
    elif model.availability_any:
        stream_av = nb.njit(lambda c: np.ones((n_alts), dtype=np.int8))
    else:
        raise NotImplementedError

    if "choice_ca" in streamers:
        _stream_ch = streamers["choice_ca"].compiled
        stream_ch = nb.njit(lambda c: np.reshape(_stream_ch(c), -1))
    elif "choice_co_code" in streamers:
        stream_ch = streamers["choice_co_code"]
    else:
        raise NotImplementedError

    if "weight_co" in streamers:
        stream_wt = streamers["weight_co"].compiled
    else:
        stream_wt = nb.njit(lambda c: np.ones((1,), dtype=float_dtype))

    @nb.njit
    def streamer(
        c,
        model_q_ca_param_scale,  # [0] float input shape=[n_q_ca_features]
        model_q_ca_param,  # [1] int input shape=[n_q_ca_features]
        model_q_ca_data,  # [2] int input shape=[n_q_ca_features]
        model_q_scale_param,  # [3] int input scalar
        model_utility_ca_param_scale,  # [4] float input shape=[n_u_ca_features]
        model_utility_ca_param,  # [5] int input shape=[n_u_ca_features]
        model_utility_ca_data,  # [6] int input shape=[n_u_ca_features]
        model_utility_co_alt,  # [ 7] int input shape=[n_co_features]
        model_utility_co_param_scale,  # [ 8] float input shape=[n_co_features]
        model_utility_co_param,  # [ 9] int input shape=[n_co_features]
        model_utility_co_data,  # [10] int input shape=[n_co_features]
        edgeslots,  # [11] int input shape=[edges, 4]
        mu_slots,  # [12] int input shape=[nests]
        start_slots,  # [13] int input shape=[nests]
        len_slots,  # [14] int input shape=[nests]
        holdfast_arr,  # [15] int8 input shape=[n_params]
        parameter_arr,  # [16] float input shape=[n_params]
        # array_ch,  # [17] float input shape=[nodes]
        # array_av,  # [18] int8 input shape=[nodes]
        # array_wt,  # [19] float input shape=[]
        # array_co,  # [20] float input shape=[n_co_vars]
        # array_ca,  # [21] float input shape=[n_alts, n_ca_vars]
        # array_ce_data,  # [22] float input shape=[n_casealts, n_ca_vars]
        # array_ce_indices,  # [23] int input shape=[n_casealts]
        # array_ce_ptr,  # [24] int input shape=[2]
        return_flags,
        # utility,  # [23] float output shape=[nodes]
        # logprob,  # [24] float output shape=[nodes]
        # probability,  # [25] float output shape=[nodes]
        # bhhh,  # [26] float output shape=[n_params, n_params]
        # d_loglike,  # [27] float output shape=[n_params]
        # loglike,  # [28] float output shape=[]
    ):
        array_ch_ = stream_ch(c)  # float input shape=[nodes]
        array_av_ = stream_av(c)  # int8 input shape=[nodes]
        array_wt = stream_wt(c)  # float input shape=[]
        array_co = stream_co(c)  # float input shape=[n_co_vars]
        array_ca = stream_ca(c)  # float input shape=[n_alts, n_ca_vars]

        n_alts = array_ca.shape[0]
        n_nodes = mu_slots.shape[0] + n_alts

        array_av = cascade_sum_1row(
            array_av_, edgeslots[:, 1], edgeslots[:, 0], n_nodes, 127
        )
        array_ch = cascade_sum_1row(
            array_ch_, edgeslots[:, 1], edgeslots[:, 0], n_nodes
        )

        # assert edgeslots.shape[1] == 4
        # upslots   = edgeslots[:,0]  # int input shape=[edges]
        # dnslots   = edgeslots[:,1]  # int input shape=[edges]
        # visit1    = edgeslots[:,2]  # int input shape=[edges]
        # allocslot = edgeslots[:,3]  # int input shape=[edges]

        assert return_flags.size == 4
        only_utility = return_flags[0]  # int8 input
        # return_probability = return_flags[1]    # bool input
        # return_grad = return_flags[2]           # bool input
        # return_bhhh = return_flags[3]           # bool input

        utility = np.zeros(n_nodes, dtype=float_dtype)
        dutility = np.zeros((utility.size, parameter_arr.size), dtype=utility.dtype)
        n_params = parameter_arr.shape[0]
        logprob = np.zeros(n_nodes, dtype=float_dtype)
        probability = np.zeros(n_nodes, dtype=float_dtype)
        bhhh = np.zeros((n_params, n_params), dtype=float_dtype)
        d_loglike = np.zeros(n_params, dtype=float_dtype)
        loglike = np.zeros(1, dtype=float_dtype)

        quantity_from_data_ca(
            model_q_ca_param_scale,  # float input shape=[n_q_ca_features]
            model_q_ca_param,  # int input shape=[n_q_ca_features]
            model_q_ca_data,  # int input shape=[n_q_ca_features]
            model_q_scale_param,  # int input scalar
            parameter_arr,  # float input shape=[n_params]
            holdfast_arr,  # float input shape=[n_params]
            array_av,  # int8 input shape=[n_nodes]
            array_ca,  # float input shape=[n_alts, n_ca_vars]
            utility[:n_alts],  # float output shape=[n_alts]
            dutility[:n_alts],
        )

        if only_utility == 3:
            if model_q_scale_param[0] >= 0:
                scale_param_value = parameter_arr[model_q_scale_param[0]]
            else:
                scale_param_value = 1.0
            utility[:n_alts] = np.exp(utility[:n_alts] / scale_param_value)
            return utility, dutility, logprob, probability, bhhh, d_loglike, loglike

        # TODO not zeros?
        array_ce_data = np.zeros((0, array_ca.shape[1]), dtype=array_ca.dtype)
        array_ce_indices = np.zeros((0,), dtype=np.int32)
        array_ce_ptr = np.zeros((2,), dtype=np.int32)

        utility_from_data_ca(
            model_utility_ca_param_scale,  # int input shape=[n_u_ca_features]
            model_utility_ca_param,  # int input shape=[n_u_ca_features]
            model_utility_ca_data,  # int input shape=[n_u_ca_features]
            parameter_arr,  # float input shape=[n_params]
            holdfast_arr,  # float input shape=[n_params]
            array_av,  # int8 input shape=[n_nodes]
            array_ca,  # float input shape=[n_alts, n_ca_vars]
            array_ce_data,  # float input shape=[n_casealts, n_ca_vars]
            array_ce_indices,  # int input shape=[n_casealts]
            array_ce_ptr,  # int input shape=[2]
            utility[:n_alts],  # float output shape=[n_alts]
            dutility[:n_alts],
        )

        utility_from_data_co(
            model_utility_co_alt,  # int input shape=[n_co_features]
            model_utility_co_param_scale,  # float input shape=[n_co_features]
            model_utility_co_param,  # int input shape=[n_co_features]
            model_utility_co_data,  # int input shape=[n_co_features]
            parameter_arr,  # float input shape=[n_params]
            holdfast_arr,  # float input shape=[n_params]
            array_av,  # int8 input shape=[n_nodes]
            array_co,  # float input shape=[n_co_vars]
            utility[:n_alts],  # float output shape=[n_alts]
            dutility[:n_alts],
        )

        if only_utility == 1:
            return utility, dutility, logprob, probability, bhhh, d_loglike, loglike

        _numba_utility_to_loglike(
            n_alts,
            edgeslots,  # int input shape=[edges, 4]
            mu_slots,  # int input shape=[nests]
            start_slots,  # int input shape=[nests]
            len_slots,  # int input shape=[nests]
            holdfast_arr,  # int8 input shape=[n_params]
            parameter_arr,  # float input shape=[n_params]
            array_ch,  # float input shape=[nodes]
            array_av,  # int8 input shape=[nodes]
            array_wt,  # float input shape=[]
            return_flags,
            dutility,
            utility,  # float output shape=[nodes]
            logprob,  # float output shape=[nodes]
            probability,  # float output shape=[nodes]
            bhhh,  # float output shape=[n_params, n_params]
            d_loglike,  # float output shape=[n_params]
            loglike,  # float output shape=[]
        )

        return utility, dutility, logprob, probability, bhhh, d_loglike, loglike

    return streamer


def init_choice_avail_summary_streamer(model):
    streamers = get_data_streamers(model, model.float_dtype)
    float_dtype = model.float_dtype
    n_alts = model.datatree.n_alts

    if "ca" in streamers:
        stream_ca = streamers["ca"]  # noqa: F841
    else:
        stream_ca = nb.njit(lambda c: np.zeros((1, n_alts, 0), dtype=float_dtype))  # noqa: F841

    if "co" in streamers:
        stream_co = streamers["co"]  # noqa: F841
    else:
        stream_co = nb.njit(lambda c: np.zeros((1, 0), dtype=float_dtype))  # noqa: F841

    if "avail_ca" in streamers:
        _stream_av_1 = streamers["avail_ca"]
        _stream_av = streamers["avail_ca"].compiled
        stream_av = nb.njit(lambda c: np.reshape(_stream_av(c), -1))
    elif model.availability_any:
        stream_av = nb.njit(lambda c: np.ones((n_alts), dtype=np.int8))
    else:
        raise NotImplementedError

    if "choice_ca" in streamers:
        _stream_ch = streamers["choice_ca"].compiled
        stream_ch = nb.njit(lambda c: np.reshape(_stream_ch(c), -1))
    elif "choice_co_code" in streamers:
        stream_ch = streamers["choice_co_code"]
    else:
        raise NotImplementedError

    if "weight_co" in streamers:
        stream_wt = streamers["weight_co"]
    else:
        stream_wt = nb.njit(lambda c: np.ones((1,), dtype=float_dtype))

    @nb.njit(parallel=False, nogil=True)
    def array_streamer(
        n_cases,
        edgeslots,  # [11] int input shape=[edges, 4]
        mu_slots,  # [12] int input shape=[nests]
        n_alts,
    ):
        n_nodes = mu_slots.shape[0] + n_alts
        total_ch = np.zeros(n_nodes, dtype=np.float32)
        total_av = np.zeros(n_nodes, dtype=np.float32)
        total_wt = np.zeros((1,), dtype=np.float32)
        eslots1 = edgeslots[:, 1]
        eslots0 = edgeslots[:, 0]
        for c in range(n_cases):
            total_av += cascade_sum_1row(stream_av(c), eslots1, eslots0, n_nodes, 127)
            n_ch = cascade_sum_1row(stream_ch(c), eslots1, eslots0, n_nodes)
            total_ch += n_ch
            total_wt += stream_wt(c) * n_ch[-1]  # float input shape=[]
        return total_ch, total_av, total_wt

    return array_streamer


@nb.njit(nogil=True, cache=True)
def _ll_streaming(
    pvals,
    c,
    _streamfunc,
    _fixed_arrays,
    _pholdfast,
    return_grad=True,
    return_bhhh=False,
):
    utility, dutility, logprob, probability, bhhh, d_loglike, loglike = _streamfunc(
        c,
        *_fixed_arrays,
        _pholdfast,
        pvals,
        np.asarray(
            [
                False,  # only_utility
                False,  # return_probability
                return_grad,
                return_bhhh,
            ],
            dtype=np.int8,
        ),
    )
    return loglike[0], d_loglike, bhhh


@nb.njit(parallel=True, nogil=True, cache=True)
def _loglike_streaming(
    pvals, start_case, stop_case, _streamfunc, _fixed_arrays, _pholdfast
):
    result = 0
    for c in nb.prange(start_case, stop_case):
        ll, _, _ = _ll_streaming(
            pvals,
            c,
            _streamfunc,
            _fixed_arrays,
            _pholdfast,
            return_grad=False,
            return_bhhh=False,
        )
        result += ll
    return result


@nb.njit(parallel=True, nogil=True, cache=True)
def _d_loglike_streaming(
    pvals, start_case, stop_case, _streamfunc, _fixed_arrays, _pholdfast
):
    result = 0
    d_result = np.zeros_like(pvals)
    for c in nb.prange(start_case, stop_case):
        ll, dll, _ = _ll_streaming(
            pvals,
            c,
            _streamfunc,
            _fixed_arrays,
            _pholdfast,
            return_grad=True,
            return_bhhh=False,
        )
        result += ll
        d_result += dll
    return result, d_result


class ModelStreamer:
    def __init__(self, model=None):
        self._model = model
        # if model is None:
        #     return
        # if not hasattr(self._model, "_streamfunc"):
        #     self._model._streamfunc = init_streamer(model)
        # _fixed_arrays = tuple(self._model._fixed_arrays)
        # _pholdfast = self._model.pholdfast
        #
        # self._choice_avail_summary_streamer = init_choice_avail_summary_streamer(
        #     self._model
        # )

        # def _loglike_func_py(pvals, c, _streamfunc, return_grad=True, return_bhhh=False):
        #     utility, dutility, logprob, probability, bhhh, d_loglike, loglike = _streamfunc(
        #         c, *_fixed_arrays, _pholdfast, pvals, np.asarray(
        #             [
        #                 False,  # only_utility
        #                 False,  # return_probability
        #                 return_grad,
        #                 return_bhhh,
        #             ],
        #             dtype=np.int8,
        #         ))
        #     return loglike[0], d_loglike, bhhh
        # self._loglike_case = _loglike_func_nb = nb.njit(_loglike_func_py)

        # def _loglike_sum(pvals, start_case, stop_case, _streamfunc):
        #     result = 0
        #     for c in nb.prange(start_case, stop_case):
        #         result += _loglike_func_nb(pvals, c, _streamfunc)
        #     return result
        # print("init _loglike_sum")
        # self._loglike_sum = nb.njit(_loglike_sum, parallel=True)

    def reset_streamer(self):
        self._model._streamfunc = init_streamer(self._model)

    def __set_name__(self, owner, name):
        self.name = name
        self.private_name = "_" + name

    def __get__(self, obj, objtype=None):
        """
        Get the value of an attribute.

        Parameters
        ----------
        obj : Any
            Instance of parent class that has `self` as a member.
        objtype : type
            Class of `obj`.
        """
        if obj is None:
            return self
        s = getattr(obj, self.private_name, None)
        if s is None:
            obj.unmangle(structure_only=True)
            s = type(self)(obj)
            setattr(obj, self.private_name, s)
        return s

    def __set__(self, instance, value):
        """
        Set the value of an attribute.

        Parameters
        ----------
        instance : Any
            Instance of parent class that has `self` as a member.
        value : Any
            Value being set.
        """
        if value is not None:
            raise NotImplementedError
        setattr(instance, self.private_name, value)
        # existing_value = self.__get__(instance)
        # if value is not None:
        #     value = str(value)
        # if existing_value != value:
        #     setattr(instance, self.private_name, value)
        #     try:
        #         instance.mangle()
        #     except AttributeError:
        #         pass

    def __delete__(self, instance):
        self.__set__(instance, None)

    def loglike(self, x=None, *, start_case=0, stop_case=-1, step_case=1, **kwargs):
        """
        Compute the log likelihood of the model.

        Parameters
        ----------
        x : array-like or dict, optional
            New values to set for the parameters before evaluating
            the log likelihood.  If given as array-like, the array must
            be a vector with length equal to the length of the
            parameter frame, and the given vector will replace
            the current values.  If given as a dictionary,
            the dictionary is used to update the parameters.
        start_case : int, default 0
            The first case to include in the log likelihood computation.
            To include all cases, start from 0 (the default).
        stop_case : int, default -1
            One past the last case to include in the log likelihood
            computation.  This is processed as usual for Python slicing
            and iterating, and negative values count backward from the
            end.  To include all cases, end at -1 (the default).
        step_case : int, default 1
            The step size of the case iterator to use in likelihood
            calculation.  This is processed as usual for Python slicing
            and iterating.  To include all cases, step by 1 (the default).

        Returns
        -------
        float
        """
        if start_case is None:
            start_case = 0
        if stop_case is None:
            stop_case = -1
        if step_case is None:
            step_case = 1
        if step_case != 1:
            raise NotImplementedError
        if stop_case < 0:
            stop_case = self._model.n_cases + 1 + stop_case
        if x is not None:
            self._model.pvals = x
        _fixed_arrays = tuple(self._model._fixed_arrays)
        _pholdfast = self._model.pholdfast
        return _loglike_streaming(
            self._model.pvals,
            start_case,
            stop_case,
            self._model._streamfunc,
            _fixed_arrays,
            _pholdfast,
        )

    def d_loglike(self, x=None, *, start_case=0, stop_case=-1, step_case=1, **kwargs):
        """
        Compute the log likelihood of the model.

        Parameters
        ----------
        x : array-like or dict, optional
            New values to set for the parameters before evaluating
            the log likelihood.  If given as array-like, the array must
            be a vector with length equal to the length of the
            parameter frame, and the given vector will replace
            the current values.  If given as a dictionary,
            the dictionary is used to update the parameters.
        start_case : int, default 0
            The first case to include in the log likelihood computation.
            To include all cases, start from 0 (the default).
        stop_case : int, default -1
            One past the last case to include in the log likelihood
            computation.  This is processed as usual for Python slicing
            and iterating, and negative values count backward from the
            end.  To include all cases, end at -1 (the default).
        step_case : int, default 1
            The step size of the case iterator to use in likelihood
            calculation.  This is processed as usual for Python slicing
            and iterating.  To include all cases, step by 1 (the default).

        Returns
        -------
        float
        """
        if start_case is None:
            start_case = 0
        if stop_case is None:
            stop_case = -1
        if step_case is None:
            step_case = 1
        if step_case != 1:
            raise NotImplementedError
        if stop_case < 0:
            stop_case = self._model.n_cases + 1 + stop_case
        if x is not None:
            self._model.pvals = x
        _fixed_arrays = tuple(self._model._fixed_arrays)
        _pholdfast = self._model.pholdfast
        return _d_loglike_streaming(
            self._model.pvals,
            start_case,
            stop_case,
            self._model._streamfunc,
            _fixed_arrays,
            _pholdfast,
        )[1]

    def choice_avail_summary(self):
        total_ch, total_av, total_wt = self._choice_avail_summary_streamer(
            self._model.n_cases,
            self._model._fixed_arrays.edge_slots,
            self._model._fixed_arrays.mu_slot,
            self._model.datatree.n_alts,
        )
        od = {}
        graph = self._model.graph
        idx = pd.Index(graph.standard_sort, name="altid")
        od["name"] = pd.Series(graph.standard_sort_names, index=idx)
        od["chosen"] = pd.Series(total_ch, index=idx)
        od["available"] = pd.Series(total_av, index=idx)
        result = pd.DataFrame(od, index=idx)
        from ..dataset.choice_avail_reporting import clean_summary

        return clean_summary(result, root_id=graph.root_id)

    def total_weight(self):
        if getattr(self, "_total_weight", None) is None:
            total_ch, total_av, total_wt = self._choice_avail_summary_streamer(
                self._model.n_cases,
                self._model._fixed_arrays.edge_slots,
                self._model._fixed_arrays.mu_slot,
                self._model.datatree.n_alts,
            )
            self._total_weight = float(total_wt)
        return self._total_weight



def example(extract='m', estimate=False):
    import larch

    larch.__version__

    d = larch.examples.MTC(format="dataset")
    d

    m = larch.Model(d, compute_engine="numba")

    from larch import PX, P, X

    m.utility_co[2] = P("ASC_SR2")
    m.utility_co[3] = P("ASC_SR3P")
    m.utility_co[4] = P("ASC_TRAN") + P("hhinc#4") * X("hhinc")
    m.utility_co[5] = P("ASC_BIKE") + P("hhinc#5") * X("hhinc")
    m.utility_co[6] = P("ASC_WALK") + P("hhinc#6") * X("hhinc")

    m.utility_ca = PX("tottime") + PX("totcost")

    m.availability_ca_var = "avail"
    m.choice_ca_var = "chose"

    m.title = "MTC Example 3, Zeroed Shared"

    m.choice_avail_summary()

    m.set_cap(20)

    assert m.compute_engine == "numba"
    if not estimate:
        if isinstance(extract, str):
            return locals()[extract]
        else:
            _locals = locals()
            return [_locals.get(i) for i in extract]
    result = m.maximize_loglike(stderr=True)
    m.calculate_parameter_covariance()
    m.loglike()

    m.parameter_summary()

    m.ordering = (
        (
            "LOS",
            "totcost",
            "tottime",
        ),
        (
            "ASCs",
            "ASC.*",
        ),
        (
            "Income",
            "hhinc.*",
        ),
    )

    m.parameter_summary()

    m.estimation_statistics()

    # TEST
    revealed_x = dict(zip(m.pnames, result.x))

    # TEST
    from pytest import approx

    expected_x = {
        "ASC_BIKE": -2.3989187974851647,
        "ASC_SR2": -2.3041966926745103,
        "ASC_SR3P": -3.7034452031908507,
        "ASC_TRAN": -0.6966955353312382,
        "ASC_WALK": -0.2269314175275228,
        "hhinc#4": -0.00486415050538345,
        "hhinc#5": -0.012431555290989793,
        "hhinc#6": -0.009365241345664176,
        "totcost": -0.004912235245649822,
        "tottime": -0.05135135046958243,
    }
    for k in expected_x:
        assert revealed_x[k] == approx(
            expected_x[k], 2e-2
        ), f"{k}, {revealed_x[k]/expected_x[k]}"
    if isinstance(extract, str):
        return locals()[extract]
    else:
        _locals = locals()
        return [_locals.get(i) for i in extract]
